from django.conf.urls import url

from . import views

urlpatterns = [
    url("^save/$", views.save, name="usermedia_save"),
    url("^delete/$", views.delete, name="usermedia_delete"),
    url("^images/$", views.images, name="usermedia_images"),
    url(
        "^save_category/$", views.save_category, name="usermedia_save_category"
    ),
]



# Generated by Django 2.2.9 on 2020-02-05 17:27
import os
import zipfile
import tempfile
from decimal import Decimal

from django.db import migrations, models
from django.core.files import File

# FW 3.1 documents can be upgraded to 3.2 without changes
# (not true for reverse conversion)
OLD_FW_DOCUMENT_VERSION = 3.1
FW_DOCUMENT_VERSION = 3.2


# from https://stackoverflow.com/questions/25738523/how-to-update-one-file-inside-zip-file-using-python
def update_revision_zip(file_field, file_name):
    # generate a temp file
    tmpfd, tmpname = tempfile.mkstemp()
    os.close(tmpfd)
    # create a temp copy of the archive without filename
    with zipfile.ZipFile(file_field.open(), "r") as zin:
        with zipfile.ZipFile(tmpname, "w") as zout:
            zout.comment = zin.comment  # preserve the comment
            for item in zin.infolist():
                if item.filename == "filetype-version":
                    zout.writestr(item, str(FW_DOCUMENT_VERSION))
                else:
                    zout.writestr(item, zin.read(item.filename))
    # replace with the temp archive
    with open(tmpname, "rb") as tmp_file:
        file_field.save(file_name, File(tmp_file))
    os.remove(tmpname)


def set_document_version(apps, schema_editor):
    Document = apps.get_model("document", "Document")
    documents = Document.objects.all().iterator()
    for document in documents:
        if document.doc_version == Decimal(str(OLD_FW_DOCUMENT_VERSION)):
            document.doc_version = FW_DOCUMENT_VERSION
            for field in document._meta.local_fields:
                if field.name == "updated":
                    field.auto_now = False
            document.save()

    DocumentTemplate = apps.get_model("document", "DocumentTemplate")
    templates = DocumentTemplate.objects.all()
    for template in templates:
        if template.doc_version == Decimal(str(OLD_FW_DOCUMENT_VERSION)):
            template.doc_version = FW_DOCUMENT_VERSION
            template.save()

    DocumentRevision = apps.get_model("document", "DocumentRevision")
    revisions = DocumentRevision.objects.all()
    for revision in revisions:
        if not revision.file_object:
            revision.delete()
            continue
        if revision.doc_version == Decimal(str(OLD_FW_DOCUMENT_VERSION)):
            revision.doc_version = FW_DOCUMENT_VERSION
            revision.save()
            # Set the version number also in the zip file.
            update_revision_zip(revision.file_object, revision.file_name)


class Migration(migrations.Migration):

    dependencies = [
        ("document", "0001_squashed_20200219"),
    ]

    operations = [
        migrations.AlterField(
            model_name="document",
            name="doc_version",
            field=models.DecimalField(
                decimal_places=1, default=3.2, max_digits=3
            ),
        ),
        migrations.AlterField(
            model_name="documentrevision",
            name="doc_version",
            field=models.DecimalField(
                decimal_places=1, default=3.2, max_digits=3
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplate",
            name="doc_version",
            field=models.DecimalField(
                decimal_places=1, default=3.2, max_digits=3
            ),
        ),
        migrations.RunPython(set_document_version),
    ]



import shutil
import os
from subprocess import call

from base.management import BaseCommand
from django.core.management import call_command
from django.conf import settings


class Command(BaseCommand):
    help = "Check CSS files with Stylelint"

    def add_arguments(self, parser):
        parser.add_argument(
            "--fix",
            action="store_true",
            dest="fix",
            default=False,
            help="Whether to also attempt to fix files",
        )

    def handle(self, *args, **options):
        self.stdout.write("Linting CSS code...")
        call_command("npm_install")
        shutil.os.chdir(settings.PROJECT_PATH)
        command_array = [
            os.path.join(
                settings.PROJECT_PATH, ".transpile/node_modules/.bin/stylelint"
            ),
            "**/*.css",
            "!static-libs/**",
            "!static-collected/**",
            "!venv/**",
            "--allow-empty-input",
            "--config",
            os.path.join(settings.SRC_PATH, "stylelint.config.js"),
        ]
        if options["fix"]:
            command_array.append("--fix")
        return_value = call(command_array)
        if return_value > 0:
            exit(return_value)



from builtins import range
from builtins import object
import re
import os
import time

from django.test import Client
from selenium import webdriver
from allauth.account.models import EmailAddress
from django.contrib.auth.hashers import make_password
from django.contrib.auth import get_user_model


class SeleniumHelper(object):
    """
    Methods for manipulating django and the browser for testing purposes.
    """

    @classmethod
    def get_drivers(cls, number, download_dir=False, user_agent=False):
        # django native clients, to be used for faster login.
        clients = []
        for i in range(number):
            clients.append(Client())
        drivers = []
        wait_time = 0
        options = webdriver.ChromeOptions()
        if download_dir:
            prefs = {
                "download.default_directory": download_dir,
                "download.prompt_for_download": False,
                "download.directory_upgrade": True,
            }
            options.add_experimental_option("prefs", prefs)
        if user_agent:
            options.add_argument("user-agent={}".format(user_agent))
        if os.getenv("CI"):
            options.binary_location = "/usr/bin/google-chrome-stable"
            options.add_argument("--headless")
            options.add_argument("--disable-gpu")
            chromedriver_filename = None
            wait_time = 13
        else:
            from chromedriver_binary import chromedriver_filename

            wait_time = 6
        for i in range(number):
            if chromedriver_filename:
                driver = webdriver.Chrome(
                    chromedriver_filename, options=options
                )
            else:
                driver = webdriver.Chrome(options=options)
            drivers.append(driver)
        for driver in drivers:
            # Set sizes of browsers so that all buttons are visible.
            driver.set_window_position(0, 0)
            driver.set_window_size(1920, 1080)
        return {"clients": clients, "drivers": drivers, "wait_time": wait_time}

    def find_urls(self, string):
        return re.findall(
            (
                "http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*(),]|"
                "(?:%[0-9a-fA-F][0-9a-fA-F]))+"
            ),
            string,
        )

    # create django data
    def create_user(
        self, username="User", email="test@example.com", passtext="p4ssw0rd"
    ):
        User = get_user_model()
        user = User.objects.create(
            username=username,
            email=email,
            password=make_password(passtext),
            is_active=True,
        )
        user.save()

        # avoid the unverified-email login trap
        EmailAddress.objects.create(
            user=user, email=email, verified=True, primary=True
        ).save()

        return user

    # drive browser
    def login_user(self, user, driver, client):
        client.force_login(user=user)
        cookie = client.cookies["sessionid"]
        if driver.current_url == "data:,":
            # To set the cookie at the right domain we load the front page.
            driver.get("%s%s" % (self.live_server_url, "/"))
        driver.add_cookie(
            {
                "name": "sessionid",
                "value": cookie.value,
                "secure": False,
                "path": "/",
            }
        )

    def leave_site(self, driver):
        driver.get("data:,")

    def wait_until_file_exists(self, path, wait_time):
        count = 0
        while not os.path.exists(path):
            time.sleep(1)
            count += 1
            if count > wait_time:
                break



from django.conf.urls import url

from . import views

urlpatterns = [
    url("^flatpage/$", views.flatpage, name="flatpage"),
    url("^configuration/$", views.configuration, name="configuration"),
]



from base.management import BaseCommand
from django.core.management import call_command


class Command(BaseCommand):
    help = "Lint JavaScript and Python files"

    def handle(self, *args, **options):
        call_command("lint_js")
        call_command("lint_css")
        call_command("lint_py")



import time
import os
from tempfile import mkdtemp

from testing.testcases import LiveTornadoTestCase
from testing.selenium_helper import SeleniumHelper
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.action_chains import ActionChains


class AdminTest(LiveTornadoTestCase, SeleniumHelper):
    fixtures = [
        "initial_documenttemplates.json",
        "initial_styles.json",
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.base_url = cls.live_server_url
        cls.download_dir = mkdtemp()
        cls.base_admin_url = f"{cls.base_url}/admin/"
        driver_data = cls.get_drivers(1, cls.download_dir)
        cls.driver = driver_data["drivers"][0]
        cls.client = driver_data["clients"][0]
        cls.driver.implicitly_wait(driver_data["wait_time"])
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        os.rmdir(cls.download_dir)
        super().tearDownClass()

    def setUp(self):
        self.verificationErrors = []
        self.accept_next_alert = True
        self.admin = self.create_user(
            username="Admin", email="admin@admin.com", passtext="password"
        )
        self.admin.is_superuser = True
        self.admin.is_staff = True
        self.admin.save()
        self.user1 = self.create_user(
            username="User1", email="user1@user.com", passtext="password"
        )
        self.user2 = self.create_user(
            username="User2", email="user2@user.com", passtext="password"
        )

    def tearDown(self):
        self.leave_site(self.driver)

    def test_maintenance(self):
        self.driver.get(self.base_admin_url)
        username = self.driver.find_element(By.ID, "id_username")
        username.send_keys("Admin")
        self.driver.find_element(By.ID, "id_password").send_keys("password")
        self.driver.find_element(By.CSS_SELECTOR, "input[type=submit]").click()
        time.sleep(2)
        self.driver.find_element(
            By.CSS_SELECTOR, "a[href='/admin/document/document/maintenance/']"
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, "#update").click()
        self.assertEqual(
            1,
            len(
                self.driver.find_elements_by_css_selector("#update[disabled]")
            ),
        )

    def test_templates(self):
        self.driver.get(self.base_admin_url)
        username = self.driver.find_element(By.ID, "id_username")
        username.send_keys("Admin")
        self.driver.find_element(By.ID, "id_password").send_keys("password")
        self.driver.find_element(By.CSS_SELECTOR, "input[type=submit]").click()
        time.sleep(2)
        self.driver.find_element(
            By.CSS_SELECTOR, "a[href='/admin/document/documenttemplate/']"
        ).click()
        template_links = self.driver.find_elements_by_css_selector(
            "#result_list tbody a"
        )
        self.assertEqual(1, len(template_links))
        self.driver.find_element(
            By.CSS_SELECTOR, "input[type=checkbox].action-select"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, "select[name=action]"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, "option[value=duplicate]"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, "button[type=submit]"
        ).click()
        template_links = self.driver.find_elements_by_css_selector(
            "#result_list tbody a"
        )
        self.assertEqual(2, len(template_links))
        self.driver.find_element(By.CSS_SELECTOR, ".field-title a").click()
        self.driver.find_element(By.CSS_SELECTOR, "input.title").click()
        ActionChains(self.driver).send_keys(Keys.BACKSPACE).send_keys(
            Keys.BACKSPACE
        ).send_keys(Keys.BACKSPACE).send_keys(Keys.BACKSPACE).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            "Special Article"
        ).perform()
        self.driver.find_element(By.CSS_SELECTOR, "input.import-id").click()
        ActionChains(self.driver).send_keys(Keys.BACKSPACE).send_keys(
            Keys.BACKSPACE
        ).send_keys(Keys.BACKSPACE).send_keys(Keys.BACKSPACE).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            Keys.BACKSPACE
        ).send_keys(
            "special-article"
        ).perform()
        self.driver.find_element(
            By.CSS_SELECTOR,
            (
                "#template-editor > table:nth-child(2) > tbody > tr > "
                "td.to-column > div.to-container > div:nth-child(5) > "
                "div.doc-part-header > ul > li > span"
            ),
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            (
                "#template-editor > table:nth-child(2) > tbody > tr > "
                "td.to-column > div.to-container > div:nth-child(5) > "
                "div.attrs > div:nth-child(28) > div.initial > div > "
                "div.ProseMirror > div > p"
            ),
        ).click()
        ActionChains(self.driver).send_keys("Initial body").perform()
        self.driver.find_element(
            By.CSS_SELECTOR,
            (
                "#template-editor > table:nth-child(2) > tbody > tr > "
                "td.to-column > div.to-container > div:nth-child(5) > "
                "div.attrs > div:nth-child(29) > div.instructions > div > "
                "div.ProseMirror > p"
            ),
        ).click()
        ActionChains(self.driver).send_keys("Body instructions").perform()
        self.driver.find_element(By.ID, "id_user").click()
        self.driver.find_element(
            By.CSS_SELECTOR, "#id_user > option:nth-child(3)"
        ).click()
        # Modify a document style
        self.driver.find_element(By.CSS_SELECTOR, ".document-style").click()
        self.driver.find_element(By.CSS_SELECTOR, "input.slug").send_keys(
            "fish"
        )
        self.driver.find_element(
            By.CSS_SELECTOR,
            "[aria-describedby=document-style-dialog] button.fw-dark",
        ).click()
        time.sleep(1)
        # Delete a document style
        self.driver.find_element(By.CSS_SELECTOR, ".document-style").click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "[aria-describedby=document-style-dialog] button.fw-orange",
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "[aria-describedby=confirmdeletion] button.fw-dark",
        ).click()
        # Download export template file
        self.driver.find_element(By.CSS_SELECTOR, ".export-template").click()
        export_template_link = self.driver.find_element(
            By.CSS_SELECTOR, ".export-template-file a"
        )
        et_file = export_template_link.get_attribute("href").split("/")[-1]
        export_template_link.click()
        time.sleep(1)
        assert os.path.isfile(os.path.join(self.download_dir, et_file))
        # Delete export template
        old_len_export_templates = len(
            self.driver.find_elements_by_css_selector(".export-template")
        )
        self.driver.find_element(
            By.CSS_SELECTOR,
            "[aria-describedby=export-template-dialog] button.fw-orange",
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "[aria-describedby=confirmdeletion] button.fw-dark",
        ).click()
        time.sleep(1)
        len_export_templates = len(
            self.driver.find_elements_by_css_selector(".export-template")
        )
        self.assertEqual(old_len_export_templates - 1, len_export_templates)
        old_len_export_templates = len_export_templates
        # Upload export template file
        self.driver.find_element(
            By.CSS_SELECTOR, ".export-template .fa-plus-circle"
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, ".fw-media-file-input")
            )
        ).send_keys(os.path.join(self.download_dir, et_file))
        time.sleep(1)
        self.driver.find_element(
            By.CSS_SELECTOR, ".ui-dialog .fw-dark"
        ).click()
        time.sleep(1)
        len_export_templates = len(
            self.driver.find_elements_by_css_selector(".export-template")
        )
        self.assertEqual(old_len_export_templates + 1, len_export_templates)
        os.remove(os.path.join(self.download_dir, et_file))
        self.driver.find_element(By.CSS_SELECTOR, "input[type=submit]").click()
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, "td.field-user").text,
            "User1",
        )
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, "th.field-title").text,
            "Special Article",
        )
        self.driver.find_element(
            By.CSS_SELECTOR, "a[href='/admin/logout/']"
        ).click()
        self.driver.get(self.base_url)
        # Logging in as User 1
        self.driver.find_element(By.ID, "id_login").send_keys("User1")
        self.driver.find_element(By.ID, "id_password").send_keys("password")
        self.driver.find_element(By.ID, "login-submit").click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, ".new_document.dropdown")
            )
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            (
                "#fw-overview-menu > li.fw-overview-menu-item.new_document."
                "dropdown > div.fw-pulldown.fw-left > ul > li:nth-child(2) > "
                "span"
            ),
        ).click()
        body_text = (
            WebDriverWait(self.driver, self.wait_time)
            .until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, ".article-body"))
            )
            .text
        )
        self.assertEqual(body_text, "Initial body")
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, ".margin-box.help").text,
            "Body instructions",
        )
        # Open share dialog
        self.driver.find_element(
            By.CSS_SELECTOR, ".header-menu:nth-child(1) > .header-nav-item"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, "li:nth-child(1) > .fw-pulldown-item"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, ".ui-dialog .fw-add-button"
        ).click()
        self.driver.find_element(By.ID, "new-contact-user-string").click()
        ActionChains(self.driver).send_keys("user2@user.com").send_keys(
            Keys.TAB
        ).send_keys(Keys.RETURN).perform()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, ".collaborator-tr .fa-caret-down")
            )
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Write"]'
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, ".ui-dialog .fw-dark"
        ).click()
        self.driver.find_element(By.ID, "close-document-top").click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.ID, "preferences-btn"))
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Log out"]'
        ).click()
        # Login as User2
        self.driver.find_element(By.ID, "id_login").send_keys("User2")
        self.driver.find_element(By.ID, "id_password").send_keys("password")
        self.driver.find_element(By.ID, "login-submit").click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Go to contacts"]'
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, ".respond-invite").click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Accept invite"]'
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Documents"]'
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, "a[href='/document/1']")
            )
        ).click()
        body_text = (
            WebDriverWait(self.driver, self.wait_time)
            .until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, ".article-body"))
            )
            .text
        )
        self.assertEqual(body_text, "Initial body")
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, ".margin-box.help").text,
            "Body instructions",
        )
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").click()
        ActionChains(self.driver).send_keys(Keys.BACKSPACE).send_keys(
            Keys.BACKSPACE
        ).send_keys(Keys.BACKSPACE).send_keys(Keys.BACKSPACE).send_keys(
            "text"
        ).perform()
        time.sleep(1)
        # Create regular copy
        old_body = self.driver.find_element(By.CSS_SELECTOR, ".article-body")
        self.driver.find_element(
            By.CSS_SELECTOR, ".header-menu:nth-child(1) > .header-nav-item"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, "li:nth-child(4) > .fw-pulldown-item"
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.staleness_of(old_body)
        )
        body_text = (
            WebDriverWait(self.driver, self.wait_time)
            .until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, ".article-body"))
            )
            .text
        )

        time.sleep(1)
        self.assertEqual(body_text, "Initial text")
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, ".margin-box.help").text,
            "Body instructions",
        )
        # Create copy with different template
        old_body = self.driver.find_element(By.CSS_SELECTOR, ".article-body")
        self.driver.find_element(
            By.CSS_SELECTOR, ".header-menu:nth-child(1) > .header-nav-item"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, "li:nth-child(5) > .fw-pulldown-item"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, ".ui-dialog select.fw-button"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, "option[value='standard-article']"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, ".ui-dialog button.fw-dark"
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.staleness_of(old_body)
        )
        body_text = (
            WebDriverWait(self.driver, self.wait_time)
            .until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, ".article-body"))
            )
            .text
        )
        # The text should still be the same, but there should be no
        # instruction boxes in this template.
        self.assertEqual(body_text, "Initial text")
        instruction_boxes = self.driver.find_elements_by_css_selector(
            ".margin-box.help"
        )
        self.assertEqual(0, len(instruction_boxes))

    def test_template_export_import(self):
        self.driver.get(self.base_admin_url)
        username = self.driver.find_element(By.ID, "id_username")
        username.send_keys("Admin")
        self.driver.find_element(By.ID, "id_password").send_keys("password")
        self.driver.find_element(By.CSS_SELECTOR, "input[type=submit]").click()
        time.sleep(2)
        self.driver.find_element(
            By.CSS_SELECTOR, "a[href='/admin/document/documenttemplate/']"
        ).click()
        # Download template
        template_links = self.driver.find_elements_by_css_selector(
            "#result_list tbody a"
        )
        self.assertEqual(1, len(template_links))
        self.driver.find_element(
            By.CSS_SELECTOR, "input[type=checkbox].action-select"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, "select[name=action]"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, "option[value=download]"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, "button[type=submit]"
        ).click()
        path = os.path.join(
            self.download_dir, "standard-article.fidustemplate"
        )
        self.wait_until_file_exists(path, self.wait_time)
        assert os.path.isfile(path)

        # Disable file dialog
        self.driver.execute_script(
            (
                "HTMLInputElement.prototype.click = function() {"
                "if(this.type !== 'file') {HTMLElement.prototype.click.call(this)}"
                "}"
            )
        )
        # Upload template again
        self.driver.find_element_by_css_selector("#upload-template").click()
        time.sleep(1)
        self.driver.find_element_by_css_selector(
            "#fidus-template-uploader"
        ).send_keys(path)
        # Check whether there now are two templates
        time.sleep(1)
        template_links = self.driver.find_elements_by_css_selector(
            "#result_list tbody a"
        )
        self.assertEqual(2, len(template_links))
        # Delete file
        os.remove(path)



from allauth.account.adapter import DefaultAccountAdapter
from allauth.utils import build_absolute_uri


class AccountAdapter(DefaultAccountAdapter):
    def get_email_confirmation_url(self, request, emailconfirmation):
        """Constructs the email confirmation (activation) url.
        Note that if you have architected your system such that email
        confirmations are sent outside of the request context `request`
        can be `None` here.
        """
        url = "/account/confirm-email/{}/".format(emailconfirmation.key)
        ret = build_absolute_uri(request, url)
        return ret

    def send_mail(self, template_prefix, email, context):
        if template_prefix == "account/email/password_reset_key":
            # We replace the password reset URL to avoid a '/api' in the URL.
            key = context["password_reset_url"].split("/")[-2]
            url = "/account/change-password/{}/".format(key)
            context["password_reset_url"] = build_absolute_uri(
                context["request"], url
            )
        return super().send_mail(template_prefix, email, context)



from os.path import dirname, isdir, join, isfile

from django.conf import settings
import django.core.management.commands.loaddata
from django.core.files.storage import default_storage
import django.core.serializers
from django.db.models import signals
from django.db.models.fields.files import FileField
from django.apps import apps


def get_apps():
    return [
        _f for _f in [a.models_module for a in apps.get_app_configs()] if _f
    ]


def get_modelclasses():
    for modelclass in apps.get_models():
        yield modelclass


def models_with_filefields():
    for modelclass in get_modelclasses():
        if any(
            isinstance(field, FileField) for field in modelclass._meta.fields
        ):
            yield modelclass


class Command(django.core.management.commands.loaddata.Command):
    def load_images_for_signal(self, sender, **kwargs):
        instance = kwargs["instance"]
        for field in sender._meta.fields:
            if not isinstance(field, FileField):
                continue
            path = getattr(instance, field.attname)
            if path is None or not path.name:
                continue
            find_file = False
            for fixture_path in self.fixture_media_paths:

                filepath = join(fixture_path, path.name)
                if isfile(filepath):
                    find_file = filepath
            if find_file is False:
                self.stderr.write(
                    ("Expected file {} doesn't exist, skipping").format(
                        path.name
                    )
                )
                continue
            with open(find_file, "rb") as f:
                default_storage.save(path.name, f)

    def handle(self, *fixture_labels, **options):
        # Hook up pre_save events for all the apps' models that have
        # FileFields.
        for modelclass in models_with_filefields():
            signals.pre_save.connect(
                self.load_images_for_signal, sender=modelclass
            )
        fixture_paths = self.find_fixture_paths()
        fixture_paths = (join(path, "media") for path in fixture_paths)
        fixture_paths = [path for path in fixture_paths if isdir(path)]
        self.fixture_media_paths = fixture_paths
        return_value = super().handle(*fixture_labels, **options)
        # Disconnect signal listeners
        for modelclass in models_with_filefields():
            signals.pre_save.disconnect(
                self.load_images_for_signal, sender=modelclass
            )
        return return_value

    def find_fixture_paths(self):
        """Return the full paths to all possible fixture directories."""
        app_module_paths = []
        for app in get_apps():
            if hasattr(app, "__path__"):
                # It's a 'models/' subpackage
                for path in app.__path__:
                    app_module_paths.append(path)
            else:
                # It's a models.py module
                app_module_paths.append(app.__file__)

        app_fixtures = [
            join(dirname(path), "fixtures") for path in app_module_paths
        ]

        return app_fixtures + list(settings.FIXTURE_DIRS)



from django.core.management.base import BaseCommand
import os
import subprocess
from tempfile import mkstemp

from django.conf import settings
from django.core.management import call_command


class Command(BaseCommand):
    help = (
        "Copy the document schema used in the frontend to a JSON file so that "
        "also the backend can make use of it."
    )

    def handle(self, *args, **options):
        self.stdout.write("Copying document schema")
        call_command("npm_install")
        json_path = os.path.join(settings.PROJECT_PATH, "static-libs/json/")
        if not os.path.exists(json_path):
            os.makedirs(json_path)
        schema_json_path = os.path.join(json_path, "schema.json")
        with open(
            os.path.join(
                settings.PROJECT_PATH, "static-transpile/js/schema_export.js"
            )
        ) as js_file:
            js = js_file.read()
        node_file, node_file_path = mkstemp()
        with open(node_file, "w") as f:
            f.write(
                "global.window = {}; "
                + "global.gettext = ()=>{}; "
                + "global.self = {};\n"
                + js
            )
        json = subprocess.check_output(["node", node_file_path]).decode(
            "utf-8"
        )
        os.unlink(node_file_path)
        with open(schema_json_path, "w") as schema_json:
            schema_json.write(json)



# See https://djangosnippets.org/snippets/771/
from django.http import HttpResponseBadRequest


def ajax_required(f):
    """
    AJAX request required decorator
    use it in your views:

    @ajax_required
    def my_view(request):
        ....

    """

    def wrap(request, *args, **kwargs):
        if not request.is_ajax():
            return HttpResponseBadRequest()
        return f(request, *args, **kwargs)

    wrap.__doc__ = f.__doc__
    wrap.__name__ = f.__name__
    return wrap



import os
from testing.testcases import LiveTornadoTestCase
from testing.selenium_helper import SeleniumHelper
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC

from django.conf import settings


class UsermediaOverviewTest(LiveTornadoTestCase, SeleniumHelper):
    @classmethod
    def setUpClass(cls):
        super(UsermediaOverviewTest, cls).setUpClass()
        cls.base_url = cls.live_server_url
        driver_data = cls.get_drivers(1)
        cls.driver = driver_data["drivers"][0]
        cls.client = driver_data["clients"][0]
        cls.driver.implicitly_wait(driver_data["wait_time"])
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        super(UsermediaOverviewTest, cls).tearDownClass()

    def setUp(self):
        self.verificationErrors = []
        self.accept_next_alert = True
        self.user = self.create_user(
            username="Yeti", email="yeti@snowman.com", passtext="otter1"
        )
        self.login_user(self.user, self.driver, self.client)

    def test_overview(self):
        driver = self.driver
        driver.get(f"{self.base_url}/usermedia/")
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'All categories'])[1]/following::button[1]"
        ).click()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[1]"
        ).click()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[1]"
        ).clear()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[1]"
        ).send_keys("landscape")
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::span[3]"
        ).click()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[2]"
        ).click()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[2]"
        ).clear()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[2]"
        ).send_keys("people")
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::span[4]"
        ).click()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[3]"
        ).click()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[3]"
        ).clear()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[3]"
        ).send_keys("scientific")
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::button[2]"
        ).click()
        driver.refresh()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'All categories'])[1]/following::button[1]"
        ).click()
        self.assertEqual(
            "landscape",
            driver.find_element_by_id("categoryTitle_1").get_attribute(
                "value"
            ),
        )
        self.assertEqual(
            "people",
            driver.find_element_by_id("categoryTitle_2").get_attribute(
                "value"
            ),
        )
        self.assertEqual(
            "scientific",
            driver.find_element_by_id("categoryTitle_3").get_attribute(
                "value"
            ),
        )
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Submit'])[1]/following::button[1]"
        ).click()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit categories'])[1]/following::button[1]"
        ).click()
        driver.find_element_by_name("title").click()
        driver.find_element_by_name("title").send_keys("An image")
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Select categories'])[1]/following::label[1]"
        ).click()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'landscape'])[1]/following::label[1]"
        ).click()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'people'])[1]/following::label[1]"
        ).click()
        # image path
        imagePath = os.path.join(
            settings.PROJECT_PATH, "usermedia/tests/uploads/image.png"
        )
        driver.find_element_by_name("image").send_keys(imagePath)
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'scientific'])[1]/following::button[1]"
        ).click()
        image_title = WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, "edit-image"))
        )
        self.assertEqual("An image", image_title.text)
        search_input = driver.find_element_by_css_selector(
            ".fw-overview-menu-item .fw-button input[type=text]"
        )
        search_input.click()
        search_input.send_keys("fish")
        self.assertEqual(
            "No images available",
            driver.find_element_by_xpath(
                "(.//*[normalize-space(text()) and normalize-space(.)="
                "'Size (px)'])[1]/following::td[1]"
            ).text,
        )
        search_input.send_keys(Keys.BACKSPACE)
        search_input.send_keys(Keys.BACKSPACE)
        search_input.send_keys(Keys.BACKSPACE)
        search_input.send_keys(Keys.BACKSPACE)
        self.assertEqual(
            "An image",
            driver.find_element_by_xpath(
                "(.//*[normalize-space(text()) and normalize-space(.)="
                "'Size (px)'])[1]/following::span[3]"
            ).text,
        )
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'PNG'])[1]/following::i[1]"
        ).click()
        driver.find_element_by_xpath(
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Confirm deletion'])[1]/following::button[2]"
        ).click()
        image_placeholder = WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, "dataTables-empty"))
        )
        self.assertEqual("No images available", image_placeholder.text)

    def tearDown(self):
        self.driver.quit()
        self.assertEqual([], self.verificationErrors)



from django.conf.urls import url

from . import views

urlpatterns = [
    url("^save/$", views.save, name="bibliography_save"),
    url("^delete/$", views.delete, name="bibliography_delete"),
    url(
        "^save_category/$",
        views.save_category,
        name="bibliopgraphy_save_category",
    ),
    url(
        "^delete_category/$",
        views.delete_category,
        name="bibliography_delete_category",
    ),
    url("^biblist/$", views.biblist, name="bibliography_biblist"),
]



from django.http import JsonResponse
from django.core.exceptions import ValidationError
from django.contrib.auth.decorators import login_required
from document.helpers.serializers import PythonWithURLSerializer
from django.utils.translation import ugettext as _
from django.views.decorators.http import require_POST

from base.decorators import ajax_required
from .models import DocumentStyle, DocumentStyleFile, ExportTemplate
from document.models import DocumentTemplate


@login_required
@require_POST
@ajax_required
def delete_document_style(request):
    response = {}
    id = int(request.POST["id"])
    style_selector = {"id": id}
    if not request.user.is_staff:
        style_selector["document_template__user"] = request.user
    document_style = DocumentStyle.objects.get(**style_selector)
    if document_style.document_template.documentstyle_set.all().count() < 2:
        # We do not delete the style if there is only
        # one left for this template
        response["errors"] = {
            "template": _("The template needs at least 1 document style")
        }
        return JsonResponse(response, status=400)
    document_style.delete()
    status = 200
    return JsonResponse(response, status=status)


@login_required
@require_POST
@ajax_required
def save_document_style(request):
    response = {}
    template_id = int(request.POST["template_id"])
    template_selector = {"id": template_id}
    if not request.user.is_staff:
        template_selector["user"] = request.user
    template = DocumentTemplate.objects.get(**template_selector)
    id = int(request.POST["id"])
    if id > 0:
        document_style = DocumentStyle.objects.get(
            id=id, document_template=template
        )
        status = 200
    else:
        document_style = DocumentStyle()
        document_style.document_template = template
        status = 201
    document_style.title = request.POST["title"]
    document_style.slug = request.POST["slug"]
    document_style.contents = request.POST["contents"]
    try:
        document_style.full_clean()
        document_style.save()
    except ValidationError as e:
        response["errors"] = e.message_dict
        return JsonResponse(response, status=400)
    deleted_files = request.POST.getlist("deleted_files[]")
    added_files = request.FILES.getlist("added_files[]")
    for file in added_files:
        dsf = DocumentStyleFile()
        dsf.file = file
        dsf.style = document_style
        dsf.save()
    for file in deleted_files:
        dsf = DocumentStyleFile.objects.filter(
            style=document_style, filename=file
        ).first()
        if dsf:
            dsf.delete()
    serializer = PythonWithURLSerializer()
    response["doc_style"] = serializer.serialize(
        [document_style],
        use_natural_foreign_keys=True,
        fields=["title", "slug", "contents", "documentstylefile_set"],
    )
    return JsonResponse(response, status=status)


@login_required
@require_POST
@ajax_required
def delete_export_template(request):
    response = {}
    id = int(request.POST["id"])
    template_selector = {"id": id}
    if not request.user.is_staff:
        template_selector["document_template__user"] = request.user
    export_template = ExportTemplate.objects.get(**template_selector)
    export_template.delete()
    status = 200
    return JsonResponse(response, status=status)


@login_required
@require_POST
@ajax_required
def save_export_template(request):
    response = {}
    template_id = int(request.POST["template_id"])
    template_selector = {"id": template_id}
    if not request.user.is_staff:
        template_selector["user"] = request.user
    template = DocumentTemplate.objects.get(**template_selector)
    id = int(request.POST["id"])
    if id > 0:
        export_template = ExportTemplate.objects.get(
            id=id, document_template=template
        )
        status = 200
    else:
        export_template = ExportTemplate()
        export_template.document_template = template
        status = 201
    export_template.template_file = request.FILES["added_file"]
    export_template.file_type = request.POST["added_file_type"]
    try:
        export_template.full_clean()
        export_template.save()
    except ValidationError as e:
        response["errors"] = e.message_dict
        return JsonResponse(response, status=400)
    serializer = PythonWithURLSerializer()
    response["export_template"] = serializer.serialize(
        [export_template], fields=["file_type", "template_file", "title"]
    )
    return JsonResponse(response, status=status)



import os
import time
import multiprocessing
from random import randrange
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import WebDriverWait, Select
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.action_chains import ActionChains
from selenium.common.exceptions import ElementClickInterceptedException
from django.conf import settings
from testing.testcases import LiveTornadoTestCase
from .editor_helper import EditorHelper


class OneUserTwoBrowsersTests(LiveTornadoTestCase, EditorHelper):
    """
    Tests in which collaboration between two browsers with the same user logged
    into both browsers.
    """

    user = None
    TEST_TEXT = "Lorem ipsum dolor sit amet."
    MULTILINE_TEST_TEXT = "Lorem ipsum\ndolor sit amet."
    fixtures = [
        "initial_documenttemplates.json",
        "initial_styles.json",
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        driver_data = cls.get_drivers(2)
        cls.driver = driver_data["drivers"][0]
        cls.driver2 = driver_data["drivers"][1]
        cls.client = driver_data["clients"][0]
        cls.client2 = driver_data["clients"][1]
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        cls.driver2.quit()
        super().tearDownClass()

    def setUp(self):
        self.user = self.create_user()
        self.login_user(self.user, self.driver, self.client)
        self.login_user(self.user, self.driver2, self.client2)
        self.doc = self.create_new_document()

    def tearDown(self):
        self.leave_site(self.driver)
        self.leave_site(self.driver2)

    def get_title(self, driver):
        # Title is child 0.
        return driver.execute_script(
            "return window.theApp.page.view.state.doc.firstChild"
            ".firstChild.textContent;"
        )

    def test_typing(self):
        """
        Test typing in collaborative mode with one user using browser windows
        with the user typing separately at small, random intervals.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        title_input = self.driver.find_element_by_class_name("article-title")
        title_input2 = self.driver2.find_element_by_class_name("article-title")

        first_part = "Here is "
        second_part = "my title"

        for i in range(8):
            title_input.send_keys(second_part[i])
            time.sleep(randrange(30, 40) / 200.0)
            title_input2.send_keys(first_part[i])
            time.sleep(randrange(30, 40) / 200.0)

        # Wait for the two editors to be synched
        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(16, len(self.get_title(self.driver)))

        self.assertEqual(
            self.get_title(self.driver2), self.get_title(self.driver)
        )

        body_input = self.driver.find_element_by_class_name("article-body")
        body_input2 = self.driver2.find_element_by_class_name("article-body")
        body_input.click()
        body_input2.click()

        body_input.click()
        body_input2.click()

        for char in self.TEST_TEXT:
            body_input.send_keys(char)
            time.sleep(randrange(30, 40) / 200.0)
            body_input2.send_keys(char)
            time.sleep(randrange(30, 40) / 200.0)

        # Wait for the two editors to be synched
        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.TEST_TEXT) * 2, len(self.get_contents(self.driver))
        )

        self.assertEqual(
            self.get_contents(self.driver2), self.get_contents(self.driver)
        )

    def test_threaded_typing(self):
        """
        Test typing in collaborative mode with one user using browser windows
        with the user typing simultaneously in two different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        title_input = self.driver.find_element_by_class_name("article-title")
        title_input2 = self.driver2.find_element_by_class_name("article-title")

        first_part = "Here is "
        second_part = "my title"

        p1 = multiprocessing.Process(
            target=self.input_text, args=(title_input, second_part)
        )
        p2 = multiprocessing.Process(
            target=self.input_text, args=(title_input2, first_part)
        )
        p1.start()
        p2.start()
        p1.join()
        p2.join()

        # Wait for the two editors to be synched
        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(16, len(self.get_title(self.driver)))

        self.assertEqual(
            self.get_title(self.driver2), self.get_title(self.driver)
        )

        body_input = self.driver.find_element_by_class_name("article-body")
        body_input2 = self.driver2.find_element_by_class_name("article-body")
        body_input.click()
        body_input2.click()

        body_input.click()
        body_input2.click()

        p1 = multiprocessing.Process(
            target=self.input_text, args=(body_input, self.TEST_TEXT)
        )
        p2 = multiprocessing.Process(
            target=self.input_text, args=(body_input2, self.TEST_TEXT)
        )
        p1.start()
        p2.start()
        p1.join()
        p2.join()

        # Wait for the two editors to be synched
        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.TEST_TEXT) * 2, len(self.get_contents(self.driver))
        )

        self.assertEqual(
            self.get_contents(self.driver2), self.get_contents(self.driver)
        )

    def make_bold(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Strong"]')
        button.click()

    def get_boldtext(self, driver):
        btext = driver.find_element_by_xpath(
            '//*[contains(@class, "article-body")]/p/strong'
        )
        return btext.text

    def test_select_and_bold(self):
        """
        Test typing in collaborative mode with one user typing and
        another user bold some part of the text in two different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 22
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text, args=(document_input, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("ProseMirror")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(25,30)")

        p2 = multiprocessing.Process(
            target=self.make_bold, args=(self.driver2,)
        )
        p2.start()
        p1.join()
        p2.join()

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(5, len(self.get_boldtext(self.driver2)))

        self.assertEqual(
            len(self.get_boldtext(self.driver)),
            len(self.get_boldtext(self.driver2)),
        )

    def make_italic(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Emphasis"]')
        button.click()

    def get_italictext(self, driver):
        itext = driver.find_element_by_xpath(
            '//*[contains(@class, "article-body")]/p/em'
        )
        return itext.text

    def test_select_and_italic(self):
        """
        Test typing in collaborative mode with one user typing and
        another user italic some part of the text in two different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 23
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text, args=(document_input, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("ProseMirror")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(25,30)")

        p2 = multiprocessing.Process(
            target=self.make_italic, args=(self.driver2,)
        )
        p2.start()
        p1.join()
        p2.join()

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(5, len(self.get_italictext(self.driver2)))

        self.assertEqual(
            len(self.get_italictext(self.driver)),
            len(self.get_italictext(self.driver2)),
        )

    def make_numberedlist(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Numbered list"]')
        button.click()

    def get_numberedlist(self, driver):
        numberedTags = driver.find_elements_by_xpath(
            '//*[contains(@class, "article-body")]//ol//li'
        )
        return numberedTags

    def test_numberedlist(self):
        """
        Test typing in collaborative mode with one user typing and
        another user use numbered list button in two different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)
        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 23
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text,
            args=(document_input, self.MULTILINE_TEST_TEXT),
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("ProseMirror")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(25,25)")

        p2 = multiprocessing.Process(
            target=self.make_numberedlist, args=(self.driver2,)
        )
        p2.start()
        p2.join()

        # Wait for the first processor to write some text and go to next line
        self.wait_for_doc_size(self.driver2, 49)

        self.driver2.execute_script("window.testCaret.setSelection(43,43)")

        p2 = multiprocessing.Process(
            target=self.make_numberedlist, args=(self.driver2,)
        )
        p2.start()

        p1.join()
        p2.join()

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.get_numberedlist(self.driver)),
            len(self.get_numberedlist(self.driver2)),
        )

        self.assertEqual(2, len(self.get_numberedlist(self.driver2)))

    def make_bulletlist(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Bullet list"]')
        button.click()

    def get_bulletlist(self, driver):
        bulletTags = driver.find_elements_by_xpath(
            '//*[contains(@class, "article-body")]//ul//li'
        )
        return bulletTags

    def test_bulletlist(self):
        """
        Test typing in collaborative mode with one user typing and
        another user use bullet list button in two different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 23
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text,
            args=(document_input, self.MULTILINE_TEST_TEXT),
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("ProseMirror")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(25,25)")

        p2 = multiprocessing.Process(
            target=self.make_bulletlist, args=(self.driver2,)
        )
        p2.start()
        p2.join()

        # Wait for the first processor to write enough text and go to next line
        self.wait_for_doc_size(self.driver2, 49)

        self.driver2.execute_script("window.testCaret.setSelection(43,43)")

        p2 = multiprocessing.Process(
            target=self.make_bulletlist, args=(self.driver2,)
        )
        p2.start()

        p1.join()
        p2.join()

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(2, len(self.get_bulletlist(self.driver2)))

        self.assertEqual(
            len(self.get_bulletlist(self.driver)),
            len(self.get_bulletlist(self.driver2)),
        )

    def make_blockquote(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Blockquote"]')
        button.click()

    def get_blockquote(self, driver):
        blockquoteTags = driver.find_elements_by_xpath(
            '//*[contains(@class, "article-body")]/blockquote'
        )
        return blockquoteTags

    def test_blockquote(self):
        """
        Test typing in collaborative mode with one user typing and
        another user use block qoute button in two different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 22
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text, args=(document_input, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 25)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("ProseMirror")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(25,25)")

        p2 = multiprocessing.Process(
            target=self.make_blockquote, args=(self.driver2,)
        )
        p2.start()
        p2.join()
        p1.join()

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(1, len(self.get_blockquote(self.driver2)))

        self.assertEqual(
            len(self.get_blockquote(self.driver)),
            len(self.get_blockquote(self.driver2)),
        )

    def addlink(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Link"]')
        button.click()

        # wait to load popup
        linktitle = WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, "link-title"))
        )
        linktitle.click()
        self.input_text(linktitle, "Test link")

        link = driver.find_element_by_class_name("link")
        self.input_text(link, "example.com")

        driver.find_element_by_css_selector("button.fw-dark").click()

    def get_link(self, driver):
        atag = driver.find_element_by_xpath(
            '//*[contains(@class, "article-body")]/p/a'
        )
        return atag.text

    def test_add_link(self):
        """
        Test typing in collaborative mode with one user typing and
        another user select some part of the text and add link
        in two different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 23
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text, args=(document_input, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("ProseMirror")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(25,30)")

        p2 = multiprocessing.Process(target=self.addlink, args=(self.driver2,))
        p2.start()
        p1.join()
        p2.join()

        self.assertEqual(5, len(self.get_link(self.driver2)))

        self.assertEqual(
            len(self.get_link(self.driver)), len(self.get_link(self.driver2))
        )

    def make_footnote(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Footnote"]')
        button.click()

        # wait for footnote to be created
        WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CLASS_NAME, "footnote-container")
            )
        )

        footnote_box = driver.find_element_by_id("footnote-box-container")
        footnote_editor = footnote_box.find_element_by_class_name(
            "ProseMirror"
        )
        footnote_editor.click()

        self.input_text(footnote_editor, "footnote Text")

    def get_footnote(self, driver):
        atag = driver.find_element_by_xpath(
            '//*[@id="footnote-box-container"]/div[2]/div[1]/p[1]'
        )
        return atag.text

    def test_footnote(self):
        """
        Test typing in collaborative mode with one user typing and
        another user add a footnote in two different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 22
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text, args=(document_input, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("ProseMirror")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(28,28)")

        p2 = multiprocessing.Process(
            target=self.make_footnote, args=(self.driver2,)
        )
        p2.start()
        p1.join()
        p2.join()

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(13, len(self.get_footnote(self.driver2)))

        self.assertEqual(
            len(self.get_footnote(self.driver)),
            len(self.get_footnote(self.driver2)),
        )

    def perform_delete_undo(self, driver):
        actions = ActionChains(driver)
        actions.send_keys(Keys.BACKSPACE)
        actions.perform()

        button = driver.find_element_by_xpath('//*[@title="Undo"]')
        button.click()

    def get_undo(self, driver):
        content = driver.find_element_by_class_name("article-body")
        return (
            content.get_attribute("innerText")
            .rstrip("\ufeff\n")
            .replace("\n", "")
        )

    def test_delete_undo(self):
        """
        Test typing in collaborative mode with one user typing and
        another user delete and undo some part of the text in two different
        threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 22
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text, args=(document_input, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("article-title")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(25,30)")

        p2 = multiprocessing.Process(
            target=self.perform_delete_undo, args=(self.driver2,)
        )
        p2.start()
        p1.join()
        p2.join()

        # Wait for the two editors to be synched
        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(self.TEST_TEXT, self.get_undo(self.driver2))

        self.assertEqual(
            self.get_undo(self.driver), self.get_undo(self.driver2)
        )

    def make_mathequation(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Math"]')
        button.click()

        # wait for load of popup
        insert_button = WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, "insert-math"))
        )

        # type formula
        math_field = WebDriverWait(driver, self.wait_time).until(
            EC.visibility_of_element_located((By.CSS_SELECTOR, ".math-field"))
        )
        math_field.click()

        # wait for keyboard
        WebDriverWait(driver, self.wait_time).until(
            EC.visibility_of_element_located(
                (
                    By.CSS_SELECTOR,
                    'div.ML__keyboard.is-visible li[data-alt-keys="="]',
                )
            )
        )
        element = WebDriverWait(driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, 'li[data-alt-keys="2"]')
            )
        )
        try:
            element.click()
        except ElementClickInterceptedException:
            time.sleep(1)
            WebDriverWait(driver, self.wait_time).until(
                EC.element_to_be_clickable(
                    (By.CSS_SELECTOR, 'li[data-alt-keys="2"]')
                )
            ).click()

        driver.find_element_by_css_selector(
            'li[data-alt-keys="x-var"]'
        ).click()
        driver.find_element_by_css_selector('li[data-alt-keys="+"]').click()
        driver.find_element_by_css_selector('li[data-alt-keys="3"]').click()
        driver.find_element_by_css_selector('li[data-alt-keys="="]').click()
        driver.find_element_by_css_selector('li[data-alt-keys="7"]').click()
        # close keyboard
        driver.find_element_by_class_name("ui-dialog-titlebar").click()
        insert_button.click()

    def get_mathequation(self, driver):
        math = driver.find_element_by_xpath('//*[@class="equation"]')

        return math.get_attribute("data-equation")

    def test_mathequation(self):
        """
        Test typing in collaborative mode with one user typing and
        another user insert math equation in middle of the text in two
        different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 23
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text, args=(document_input, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 30)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("ProseMirror")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(28,28)")

        p2 = multiprocessing.Process(
            target=self.make_mathequation, args=(self.driver2,)
        )
        p2.start()
        p1.join()
        p2.join()

        # Wait for the two editors to be synched
        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.get_mathequation(self.driver)),
            len(self.get_mathequation(self.driver2)),
        )

        self.assertEqual(6, len(self.get_mathequation(self.driver2)))

    def add_comment(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Comment"]')
        button.click()

        textArea = WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, ".ProseMirror-wrapper .ProseMirror")
            )
        )
        textArea.click()

        self.input_text(textArea, "My comment")

        driver.find_element_by_css_selector(
            "div#comment-editor button.submit"
        ).click()

    def get_comment(self, driver):
        comment = driver.find_element_by_class_name("comment-text-wrapper")

        return comment.text

    def test_comment(self):
        """
        Test typing in collaborative mode with one user typing and
        another user add some comment in middle of the text in two different
        threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 22
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text, args=(document_input, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 33)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("ProseMirror")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(25,30)")

        p2 = multiprocessing.Process(
            target=self.add_comment, args=(self.driver2,)
        )
        p2.start()
        p1.join()
        p2.join()

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(10, len(self.get_comment(self.driver2)))

        self.assertEqual(
            len(self.get_comment(self.driver)),
            len(self.get_comment(self.driver2)),
        )

        # Change comment
        self.driver.find_element_by_css_selector(
            ".margin-box.comment .show-marginbox-options"
        ).click()
        self.driver.find_element_by_css_selector(".edit-comment").click()
        self.driver.find_element_by_css_selector(
            "#comment-editor p"
        ).send_keys("MODIFICATION")
        self.driver.find_element_by_css_selector(".comment-is-major").click()
        self.driver.find_element_by_css_selector(
            "#comment-editor .submit"
        ).click()
        WebDriverWait(self.driver2, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, ".margin-box.comment-is-major-bgc")
            )
        )
        self.assertEqual(22, len(self.get_comment(self.driver2)))
        self.assertEqual(
            len(self.get_comment(self.driver)),
            len(self.get_comment(self.driver2)),
        )

        # Add comment answer
        self.driver2.find_element_by_css_selector(
            ".margin-box.comment"
        ).click()
        self.driver2.find_element_by_css_selector(
            "#answer-editor .ProseMirror"
        ).send_keys("My answer")
        self.driver2.find_element_by_css_selector(
            ".comment-answer .submit"
        ).click()
        self.driver.find_element_by_css_selector(".margin-box.comment").click()
        comment_answer = WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, ".comment-answer .comment-text-wrapper p")
            )
        )
        assert comment_answer.text == "My answer"

        # Modify comment answer
        comment_answer.click()
        self.driver.find_element_by_css_selector(
            ".comment-answer .show-marginbox-options"
        ).click()
        self.driver.find_element_by_css_selector(
            ".edit-comment-answer"
        ).click()
        self.driver.find_element_by_css_selector("#answer-editor p").send_keys(
            "\nMODIFICATION"
        )
        self.driver.find_element_by_css_selector(
            ".comment-answer .submit"
        ).click()
        answer_addition = WebDriverWait(self.driver2, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, ".comment-answer .comment-text-wrapper p+p")
            )
        )
        assert answer_addition.text == "MODIFICATION"

        # Delete comment answer
        answer_addition.click()
        self.driver2.find_element_by_css_selector(
            ".comment-answer .show-marginbox-options"
        ).click()
        self.driver2.find_element_by_css_selector(
            ".delete-comment-answer"
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.invisibility_of_element_located(
                (By.CSS_SELECTOR, ".comment-answer.collapse")
            )
        )

        # Delete comment
        self.driver.find_element_by_css_selector(".margin-box.comment").click()
        self.driver.find_element_by_css_selector(
            ".margin-box.comment .show-marginbox-options"
        ).click()
        self.driver.find_element_by_css_selector(".delete-comment").click()
        WebDriverWait(self.driver2, self.wait_time).until(
            EC.invisibility_of_element_located(
                (By.CSS_SELECTOR, ".margin-box.comment")
            )
        )

    def add_figure(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Figure"]')
        button.click()
        time.sleep(1)
        # click on 'Insert image' button
        WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.XPATH, '//*[normalize-space()="Insert image"]')
            )
        ).click()

        upload_button = WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.XPATH, '//*[normalize-space()="Add new image"]')
            )
        )

        upload_button.click()

        # image path
        image_path = os.path.join(
            settings.PROJECT_PATH, "document/tests/uploads/image.png"
        )

        # inorder to select the image we send the image path in the
        # LOCAL MACHINE to the input tag
        upload_image_url = WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.XPATH, '//*[@id="editimage"]/div[1]/input[2]')
            )
        )
        upload_image_url.send_keys(image_path)

        # click on 'Upload' button
        driver.find_element_by_xpath(
            '//*[contains(@class, "ui-button") and normalize-space()="Upload"]'
        ).click()

        # click on 'Use image' button
        WebDriverWait(driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, ".fw-data-table i.fa-check")
            )
        )

        driver.find_element_by_xpath(
            '//*[normalize-space()="Use image"]'
        ).click()

        # click on 'Insert' button
        driver.find_element_by_css_selector("button.fw-dark").click()

        caption = WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, "div.article-body figure figcaption")
            )
        )
        caption.click()
        self.input_text(caption, "My figure")

    def get_image(self, driver):
        figure = driver.find_element_by_css_selector("div.article-body figure")
        image = figure.find_elements_by_tag_name("img")

        return image

    def get_caption(self, driver):
        caption = driver.find_element_by_css_selector(
            "div.article-body figure figcaption"
        )

        return caption.text

    def test_insertFigure(self):
        """
        Test typing in collaborative mode with one user typing and
        another user insert figure middle of the text in two different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)

        document_input = self.driver.find_element_by_class_name("ProseMirror")

        # Total: 23
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text, args=(document_input, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("ProseMirror")
        content.click()

        self.driver2.execute_script("window.testCaret.setSelection(28,28)")

        p2 = multiprocessing.Process(
            target=self.add_figure, args=(self.driver2,)
        )
        p2.start()
        p1.join()
        p2.join()

        # Wait for the two editors to be synched
        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(1, len(self.get_image(self.driver2)))

        self.assertEqual(
            len(self.get_image(self.driver)), len(self.get_image(self.driver2))
        )

        self.assertEqual(9, len(self.get_caption(self.driver2)))
        self.assertEqual(
            len(self.get_caption(self.driver)),
            len(self.get_caption(self.driver2)),
        )

        # Check if image is still there after reload
        self.driver.refresh()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, "div.article-body figure")
            )
        )
        self.assertEqual(1, len(self.get_image(self.driver)))

    def add_citation(self, driver):
        button = driver.find_element_by_xpath('//*[@title="Cite"]')
        button.click()

        # click on 'Register new source' button
        register_new_source = WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CLASS_NAME, "register-new-bib-source")
            )
        )
        register_new_source.click()

        # select source
        WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located((By.ID, "select-bibtype"))
        )

        # click on article
        Select(
            driver.find_element_by_id("select-bibtype")
        ).select_by_visible_text("Article")

        # fill the values
        title_of_publication = driver.find_element_by_css_selector(
            ".journaltitle .ProseMirror"
        )
        title_of_publication.click()
        title_of_publication.send_keys("My publication title")

        title = driver.find_element_by_css_selector(".title .ProseMirror")
        title.click()
        title.send_keys("My title")

        author_firstName = driver.find_element_by_css_selector(
            ".author .given .ProseMirror"
        )
        author_firstName.click()
        author_firstName.send_keys("John")

        author_lastName = driver.find_element_by_css_selector(
            ".family .ProseMirror"
        )
        author_lastName.click()
        author_lastName.send_keys("Doe")

        publication_date = driver.find_element_by_css_selector(".date .date")
        publication_date.click()
        publication_date.send_keys("2012")

        # click on Submit button
        driver.find_element_by_xpath(
            '//*[contains(@class, "ui-button") and normalize-space()="Submit"]'
        ).click()

        # Wait for source to be listed
        WebDriverWait(driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (
                    By.CSS_SELECTOR,
                    "#selected-cite-source-table .selected-source",
                )
            )
        )

        # click on Insert button
        driver.find_element_by_css_selector(".insert-citation").click()

    def get_citation_within_text(self, driver):
        cite_within_doc = driver.find_element_by_css_selector(
            "div.article-body span.citation"
        )
        return cite_within_doc.text

    def get_citation_bib(self, driver):
        cite_bib = driver.find_element_by_class_name("article-bibliography")
        return cite_bib.text

    def test_citation(self):
        """
        Test typing in collaborative mode with one user typing and
        another user add cite in two different threads.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        document_input = self.driver.find_element_by_class_name("article-body")

        # Total: 22
        self.driver.execute_script("window.testCaret.setSelection(25,25)")

        p1 = multiprocessing.Process(
            target=self.input_text, args=(document_input, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # without clicking on content the buttons will not work
        content = self.driver2.find_element_by_class_name("article-body")
        content.click()

        p2 = multiprocessing.Process(
            target=self.add_citation, args=(self.driver2,)
        )
        p2.start()
        p1.join()
        p2.join()

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(11, len(self.get_citation_within_text(self.driver2)))

        self.assertEqual(
            len(self.get_citation_within_text(self.driver)),
            len(self.get_citation_within_text(self.driver2)),
        )

        self.assertEqual(63, len(self.get_citation_bib(self.driver)))

        self.assertEqual(
            len(self.get_citation_bib(self.driver)),
            len(self.get_citation_bib(self.driver2)),
        )

        # We delete the citation again
        ActionChains(self.driver).double_click(
            self.driver.find_element_by_css_selector(
                "div.article-body span.citation"
            )
        ).send_keys(Keys.BACKSPACE).perform()

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(
                self.driver2.find_elements_by_css_selector(
                    "div.article-body span.citation"
                )
            ),
            0,
        )

        self.assertEqual(0, len(self.get_citation_bib(self.driver)))

        self.assertEqual(
            len(self.get_citation_bib(self.driver)),
            len(self.get_citation_bib(self.driver2)),
        )

    def test_chat(self):
        "Test whether chat messages can be delivered"
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)
        self.driver.find_element_by_id("messageform").click()
        actions = ActionChains(self.driver)
        actions.send_keys("hello\n")
        actions.perform()
        message_body = WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, "message-body"))
        )
        assert message_body.text == "hello"

        message_body2 = WebDriverWait(self.driver2, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, "message-body"))
        )
        assert message_body2.text == "hello"

    def test_modify_path(self):
        "Test whether changing the path in one editor changes it in the other"
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)
        self.driver.find_element(By.CSS_SELECTOR, ".article-title").click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-title").send_keys(
            "Test Article"
        )
        self.wait_for_doc_sync(self.driver, self.driver2)
        self.assertEqual(
            self.driver2.find_element(By.CSS_SELECTOR, "#document-title").text,
            "Test Article",
        )
        # Delete existing path and insert just path
        self.driver2.find_element(By.CSS_SELECTOR, "#document-title").click()
        self.driver2.find_element(
            By.CSS_SELECTOR, "#document-title"
        ).send_keys(Keys.CONTROL, "a")
        self.driver2.find_element(
            By.CSS_SELECTOR, "#document-title"
        ).send_keys(Keys.DELETE)
        self.driver2.find_element(
            By.CSS_SELECTOR, "#document-title"
        ).send_keys("/Reports/2019/")
        self.driver2.find_element(By.CSS_SELECTOR, ".article-title").click()
        self.wait_for_doc_sync(self.driver, self.driver2)
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, "#document-title").text,
            "/Reports/2019/Test Article",
        )
        # Delete existing path and insert path including filename
        self.driver.find_element(By.CSS_SELECTOR, "#document-title").click()
        self.driver.find_element(By.CSS_SELECTOR, "#document-title").send_keys(
            Keys.CONTROL, "a"
        )
        self.driver.find_element(By.CSS_SELECTOR, "#document-title").send_keys(
            Keys.DELETE
        )
        self.driver.find_element(By.CSS_SELECTOR, "#document-title").send_keys(
            "/Reports/2019/Report 23"
        )
        self.driver.find_element(By.CSS_SELECTOR, ".article-title").click()
        self.wait_for_doc_sync(self.driver, self.driver2)
        self.assertEqual(
            self.driver2.find_element(By.CSS_SELECTOR, "#document-title").text,
            "/Reports/2019/Report 23",
        )
        # Reload page and check if path is still the same.
        self.driver.refresh()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, "#document-title")
            )
        )
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, "#document-title").text,
            "/Reports/2019/Report 23",
        )
        # Delete entire path to obtain article title again.
        self.driver2.find_element(By.CSS_SELECTOR, "#document-title").click()
        self.driver2.find_element(
            By.CSS_SELECTOR, "#document-title"
        ).send_keys(Keys.CONTROL, "a")
        self.driver2.find_element(
            By.CSS_SELECTOR, "#document-title"
        ).send_keys(Keys.DELETE)
        self.driver2.find_element(By.CSS_SELECTOR, ".article-title").click()
        self.wait_for_doc_sync(self.driver, self.driver2)
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, "#document-title").text,
            "Test Article",
        )



# Generated by Django 3.1.4 on 2021-04-11 12:40

from django.conf import settings
from django.db import migrations, models


def change_user_type(apps, schema_editor):
    ContentType = apps.get_model("contenttypes", "ContentType")
    ct = ContentType.objects.filter(app_label="auth", model="user").first()
    if ct:
        ct.app_label = "user"
        ct.save()


class Migration(migrations.Migration):

    dependencies = [
        ("user", "0002_loginas"),
    ]

    operations = [
        migrations.RunPython(change_user_type),
        migrations.RemoveField(
            model_name="userprofile",
            name="user",
        ),
        migrations.AlterModelOptions(
            name="user",
            options={"verbose_name": "user", "verbose_name_plural": "users"},
        ),
        migrations.AddField(
            model_name="user",
            name="contacts",
            field=models.ManyToManyField(
                related_name="_user_contacts_+", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AlterModelTable(
            name="user",
            table=None,
        ),
        migrations.DeleteModel(
            name="UserProfile",
        ),
    ]



import os

from importlib import import_module

from django.conf.urls import include, url
from django.http import HttpResponse
from django.contrib import admin
from django.urls import path
from django.conf import settings
from django.views.i18n import JavaScriptCatalog
from django.views.static import serve as static_serve

from .views import app as app_view, api_404 as api_404_view
from .views import admin_console as admin_console_view
from .views import manifest_json


admin.site.site_header = settings.ADMIN_SITE_HEADER
admin.site.site_title = settings.ADMIN_SITE_TITLE
admin.site.index_title = settings.ADMIN_INDEX_TITLE
admin.site.index_template = "admin/overview.html"
admin_site_urls = (
    [
        url(
            "console/",
            admin.site.admin_view(admin_console_view, cacheable=True),
            name="admin_console",
        ),
    ]
    + admin.site.urls[0],
    admin.site.urls[1],
    admin.site.urls[2],
)
# Django URLs -- Notice that these are only consulted after the
# tornado_url_list found in base/servers/tornado_django_hybrid.py
urlpatterns = [
    url(
        "^robots.txt$",
        lambda r: HttpResponse(
            "User-agent: *\nDisallow: /document/\nDisallow: /bibliography/",
            mimetype="text/plain",
        ),
    ),
    url("^manifest.json$", manifest_json, name="manifest_json"),
    url(
        "^sw.js$",
        static_serve,
        {
            "document_root": os.path.join(
                settings.PROJECT_PATH, "static-transpile/js"
            ),
            "path": "sw.js",
        },
    ),
    # I18n manual language switcher
    url("^api/i18n/", include("django.conf.urls.i18n")),
    # I18n Javascript translations
    url(
        r"^api/jsi18n/$",
        JavaScriptCatalog.as_view(),
        name="javascript-catalog",
    ),
    # Admin interface
    path("admin/", admin_site_urls),
    # Login as other user
    path("admin/", include("loginas.urls")),
]

for app in settings.INSTALLED_APPS:
    try:
        _module = import_module("%s.urls" % app)
    except ImportError:
        pass
    else:
        app_name = app.rsplit(".", 1).pop()
        urlpatterns += [url("^api/%s/" % app_name, include("%s.urls" % app))]

if hasattr(settings, "EXTRA_URLS"):
    for extra_url in settings.EXTRA_URLS:
        urlpatterns += [
            url(extra_url[0], include(extra_url[1])),
        ]
if not settings.DEBUG:
    urlpatterns += [
        url("^api/.*", api_404_view, name="api_404"),
    ]

urlpatterns += [
    url("^.*/$", app_view, name="app"),
    url("^$", app_view, name="app"),
]



import os
from base.management import BaseCommand
from django.core.management import call_command
from django.conf import settings
from django.contrib.flatpages.models import FlatPage

from document.models import DocumentTemplate
from style.models import DocumentStyle


class Command(BaseCommand):
    help = "Setup Fidus Writer installation."

    def add_arguments(self, parser):
        parser.add_argument(
            "--restart",
            action="store_true",
            dest="restart",
            default=False,
            help="Flush database before initialization.",
        )
        parser.add_argument(
            "--no-static",
            action="store_false",
            dest="static",
            help="Do not collect static files.",
        )
        parser.add_argument(
            "--no-force-transpile",
            action="store_false",
            dest="force_transpile",
            help="Force a transpile if nothing has changed.",
        )

    def handle(self, *args, **options):
        if options["force_transpile"]:
            force_transpile = True
        else:
            force_transpile = False
        if options["restart"]:
            call_command("flush")
            call_command("migrate", fake=True)
        else:
            call_command("migrate")
        if DocumentTemplate.objects.count() == 0:
            call_command(
                "loaddata",
                os.path.join(
                    settings.SRC_PATH,
                    "document/fixtures/initial_documenttemplates.json",
                ),
            )
        if DocumentStyle.objects.count() == 0:
            call_command(
                "loaddata",
                os.path.join(
                    settings.SRC_PATH, "style/fixtures/initial_styles.json"
                ),
            )
        if FlatPage.objects.count() == 0:
            call_command(
                "loaddata",
                os.path.join(
                    settings.SRC_PATH, "base/fixtures/initial_terms.json"
                ),
            )
        if os.environ.get("NO_COMPILEMESSAGES") == "true" or (
            os.path.isfile(
                os.path.join(
                    settings.SRC_PATH, "locale/es/LC_MESSAGES/django.mo"
                )
            )
            and os.path.getmtime(
                os.path.join(
                    settings.SRC_PATH, "locale/es/LC_MESSAGES/django.mo"
                )
            )
            > os.path.getmtime(
                os.path.join(
                    settings.SRC_PATH, "locale/es/LC_MESSAGES/django.po"
                )
            )
        ):
            pass
        else:
            call_command("compilemessages")
        call_command("transpile", force=force_transpile)
        if options["static"] and not settings.DEBUG:
            call_command("collectstatic", interactive=False)



from django.apps import AppConfig


class UserTemplateManagerConfig(AppConfig):
    name = "user_template_manager"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from document.models import DocumentTemplate

        # End user managers document templates, so do not attempt to delete
        # them automatically when not in use.
        DocumentTemplate.auto_delete = False



from django.contrib import admin

from . import models


class UserAdmin(admin.ModelAdmin):
    pass


admin.site.register(models.User, UserAdmin)


class UserInviteAdmin(admin.ModelAdmin):
    pass


admin.site.register(models.UserInvite, UserInviteAdmin)



from selenium.webdriver.common.by import By
from testing.testcases import LiveTornadoTestCase
from testing.selenium_helper import SeleniumHelper


class AccessTest(LiveTornadoTestCase, SeleniumHelper):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.base_url = cls.live_server_url
        cls.wait_time = 10

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_ie11(self):
        driver_data = self.get_drivers(
            1,
            user_agent=(
                "Mozilla/5.0 (compatible, MSIE 11, Windows NT 6.3; "
                "Trident/7.0; rv:11.0) like Gecko"
            ),
        )
        driver = driver_data["drivers"][0]
        driver.implicitly_wait(driver_data["wait_time"])
        driver.get(self.base_url + "/")
        h1 = driver.find_element(By.CSS_SELECTOR, "h1")
        self.assertEqual(h1.text, "Browser not supported")
        driver.quit()

    def test_robots(self):
        driver_data = self.get_drivers(
            1,
            user_agent=(
                "Mozilla/5.0 (compatible; Googlebot/2.1; "
                "+http://www.google.com/bot.html)"
            ),
        )
        driver = driver_data["drivers"][0]
        driver.implicitly_wait(driver_data["wait_time"])
        driver.get(self.base_url + "/robots.txt")
        body = driver.find_element(By.CSS_SELECTOR, "body")
        self.assertEqual(body.text, "User-agent: * Disallow: /* Allow: /$")
        driver.quit()

    def test_hello(self):
        driver_data = self.get_drivers(1)
        driver = driver_data["drivers"][0]
        driver.implicitly_wait(driver_data["wait_time"])
        driver.get(self.base_url + "/hello-tornado")
        body = driver.find_element(By.CSS_SELECTOR, "body")
        self.assertEqual(body.text, "Hello from tornado")
        driver.quit()



import time
from urllib.parse import urlparse

from testing.testcases import LiveTornadoTestCase
from testing.selenium_helper import SeleniumHelper
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC


class PathTest(LiveTornadoTestCase, SeleniumHelper):
    fixtures = [
        "initial_documenttemplates.json",
        "initial_styles.json",
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.base_url = cls.live_server_url
        driver_data = cls.get_drivers(1)
        cls.driver = driver_data["drivers"][0]
        cls.client = driver_data["clients"][0]
        cls.driver.implicitly_wait(driver_data["wait_time"])
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        super().tearDownClass()

    def setUp(self):
        self.verificationErrors = []
        self.accept_next_alert = True
        self.user1 = self.create_user(
            username="Yeti", email="yeti@snowman.com", passtext="otter"
        )

    def tearDown(self):
        self.leave_site(self.driver)

    def test_move_document(self):
        self.driver.get(self.base_url)
        self.driver.find_element(By.ID, "id_login").send_keys("Yeti")
        self.driver.find_element(By.ID, "id_password").send_keys("otter")
        self.driver.find_element(By.ID, "login-submit").click()
        # Enter editor
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, ".new_document button")
            )
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, "editor-toolbar"))
        )
        self.driver.find_element(By.CSS_SELECTOR, ".article-title").click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-title").send_keys(
            "Test"
        )
        time.sleep(1)
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, "#document-title").text,
            "Test",
        )
        self.driver.find_element(By.CSS_SELECTOR, "#document-title").click()
        self.driver.find_element(By.CSS_SELECTOR, "#document-title").click()
        self.driver.find_element(By.CSS_SELECTOR, "#document-title").send_keys(
            Keys.CONTROL, "a"
        )
        self.driver.find_element(By.CSS_SELECTOR, "#document-title").send_keys(
            Keys.DELETE
        )
        self.driver.find_element(By.CSS_SELECTOR, "#document-title").send_keys(
            "/Reports/2019/Report 23"
        )
        # Exit to overview page
        self.driver.find_element(
            By.CSS_SELECTOR, "#close-document-top"
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.ID, "preferences-btn"))
        )
        self.assertEqual(
            urlparse(self.driver.current_url).path, "/documents/Reports/2019/"
        )
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, ".fw-contents h1").text,
            "/Reports/2019/",
        )
        # Create new folder 'February' and enter
        self.driver.find_element(
            By.CSS_SELECTOR, 'button[title="Create new folder"]'
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, "#new-folder-name").click()
        self.driver.find_element(
            By.CSS_SELECTOR, "#new-folder-name"
        ).send_keys("February")
        self.driver.find_element(By.CSS_SELECTOR, "button.fw-dark").click()
        time.sleep(1)
        # Enter editor
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, ".new_document button")
            )
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, "editor-toolbar"))
        )
        self.driver.find_element(By.CSS_SELECTOR, ".article-title").click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-title").send_keys(
            "February Doc"
        )
        time.sleep(1)
        self.assertEqual(
            self.driver.find_element(By.CSS_SELECTOR, "#document-title").text,
            "/Reports/2019/February/February Doc",
        )
        # Exit to overview page
        self.driver.find_element(
            By.CSS_SELECTOR, "#close-document-top"
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.ID, "preferences-btn"))
        )
        self.assertEqual(
            urlparse(self.driver.current_url).path,
            "/documents/Reports/2019/February/",
        )
        documents = self.driver.find_elements_by_css_selector(
            ".fw-contents tbody tr a.fw-data-table-title"
        )
        self.assertEqual(len(documents), 2)
        self.assertEqual(documents[0].text, "..")
        self.assertEqual(documents[1].text, "February Doc")
        # Go up one folder
        documents[0].click()
        time.sleep(1)
        self.assertEqual(
            urlparse(self.driver.current_url).path, "/documents/Reports/2019/"
        )
        documents = self.driver.find_elements_by_css_selector(
            ".fw-contents tbody tr a.fw-data-table-title"
        )
        self.assertEqual(len(documents), 3)
        self.assertEqual(documents[0].text, "..")
        self.assertEqual(documents[1].text, "February")
        self.assertEqual(documents[2].text, "Report 23")
        # Move initial document via overview page
        self.driver.find_element_by_css_selector(
            "tr:nth-child(3) > td > label"
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, ".dt-bulk-dropdown"))
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Move selected"]'
        ).click()
        time.sleep(1)
        self.assertEqual(
            self.driver.find_element(
                By.CSS_SELECTOR, 'input[placeholder="Insert path"]'
            ).get_attribute("value"),
            "/Reports/2019/Report 23",
        )
        self.driver.find_element(
            By.CSS_SELECTOR, "#move-dialog i.fa-plus-square"
        ).click()
        time.sleep(1)
        self.driver.find_element(
            By.CSS_SELECTOR, "#move-dialog i.fa-plus-square"
        ).click()
        february_folder = self.driver.find_element(
            By.CSS_SELECTOR,
            (
                "#move-dialog .file-selector .folder-content .folder-content "
                ".folder-content .folder-name"
            ),
        )
        self.assertEqual(february_folder.text.strip(), "February")
        february_folder.click()
        time.sleep(1)
        self.assertEqual(
            self.driver.find_element(
                By.CSS_SELECTOR, 'input[placeholder="Insert path"]'
            ).get_attribute("value"),
            "/Reports/2019/February/Report 23",
        )
        self.driver.find_element(
            By.XPATH,
            '//*[contains(@class, "fw-dark") and normalize-space()="Submit"]',
        ).click()
        time.sleep(1)
        # Document should be gone as it is moved into the February subfolder
        documents = self.driver.find_elements_by_css_selector(
            ".fw-contents tbody tr a.fw-data-table-title"
        )
        self.assertEqual(len(documents), 2)
        self.assertEqual(documents[0].text, "..")
        self.assertEqual(documents[1].text, "February")
        # We enter the February folder. There should be two files now
        documents[1].click()
        time.sleep(1)
        documents = self.driver.find_elements_by_css_selector(
            ".fw-contents tbody tr a.fw-data-table-title"
        )
        self.assertEqual(len(documents), 3)
        self.assertEqual(documents[0].text, "..")
        self.assertEqual(documents[1].text, "Report 23")
        self.assertEqual(documents[2].text, "February Doc")
        # Move both docs to a new folder
        self.driver.find_element_by_css_selector(
            "tr:nth-child(2) > td > label"
        ).click()
        self.driver.find_element_by_css_selector(
            "tr:nth-child(3) > td > label"
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, ".dt-bulk-dropdown"))
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Move selected"]'
        ).click()
        time.sleep(1)
        self.assertEqual(
            self.driver.find_element(
                By.CSS_SELECTOR, 'input[placeholder="Insert path"]'
            ).get_attribute("value"),
            "/Reports/2019/February/",
        )
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="New folder"]'
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, "#new-folder-name").click()
        self.driver.find_element(
            By.CSS_SELECTOR, "#new-folder-name"
        ).send_keys("Documents")
        self.driver.find_element(
            By.XPATH,
            (
                '//*[contains(@class, "fw-dark") and '
                'normalize-space()="Create folder"]'
            ),
        ).click()
        time.sleep(1)
        self.assertEqual(
            self.driver.find_element(
                By.CSS_SELECTOR, 'input[placeholder="Insert path"]'
            ).get_attribute("value"),
            "/Documents/",
        )
        self.driver.find_element(
            By.XPATH,
            '//*[contains(@class, "fw-dark") and normalize-space()="Submit"]',
        ).click()
        time.sleep(1)
        # Documents should be gone as it is moved into the February subfolder
        documents = self.driver.find_elements_by_css_selector(
            ".fw-contents tbody tr a.fw-data-table-title"
        )
        self.assertEqual(len(documents), 1)
        self.assertEqual(documents[0].text, "..")
        documents[0].click()
        time.sleep(1)
        # Also the 2019 folder should be empty
        documents = self.driver.find_elements_by_css_selector(
            ".fw-contents tbody tr a.fw-data-table-title"
        )
        self.assertEqual(len(documents), 1)
        self.assertEqual(documents[0].text, "..")
        documents[0].click()
        time.sleep(1)
        # Also the Reports folder should be empty
        documents = self.driver.find_elements_by_css_selector(
            ".fw-contents tbody tr a.fw-data-table-title"
        )
        self.assertEqual(len(documents), 1)
        self.assertEqual(documents[0].text, "..")
        documents[0].click()
        time.sleep(1)
        # There should be just one folder in the top folder.
        # The Reports folder should have been auto-deleted.
        documents = self.driver.find_elements_by_css_selector(
            ".fw-contents tbody tr a.fw-data-table-title"
        )
        self.assertEqual(len(documents), 1)
        self.assertEqual(documents[0].text, "Documents")
        documents[0].click()
        time.sleep(1)
        # There should be two docs in the Documents folder
        documents = self.driver.find_elements_by_css_selector(
            ".fw-contents tbody tr a.fw-data-table-title"
        )
        self.assertEqual(len(documents), 3)
        self.assertEqual(documents[1].text, "Report 23")
        self.assertEqual(documents[2].text, "February Doc")



# Generated by Django 2.2.9 on 2020-02-05 20:40

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("usermedia", "0001_squashed_0009_auto_20170908_0953"),
    ]

    operations = [
        migrations.AlterField(
            model_name="image",
            name="uploader",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="image_uploader",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



import os
import tempfile
import shutil
import urllib

from django.core.management.commands import makemessages
from django.core.management import call_command
from django.core.management.utils import popen_wrapper

from base.management import BaseCommand

# This makes makemessages create both translations for Python and JavaScript
# code in one go.
#
# Additionally, it works around xgettext not being able to work with ES2015
# template strings [1] by transpiling the JavaScript files in a special way to
# replace all template strings.
#
# [1] https://savannah.gnu.org/bugs/?50920


class Command(makemessages.Command, BaseCommand):
    def handle(self, *args, **options):
        call_command("transpile")
        options["ignore_patterns"] += [
            "venv",
            ".direnv",
            "node_modules",
            "static-transpile",
        ]
        options["domain"] = "django"
        super().handle(*args, **options)
        options["domain"] = "djangojs"
        self.temp_dir_out = tempfile.mkdtemp()
        self.temp_dir_in = tempfile.mkdtemp()
        super().handle(*args, **options)
        shutil.rmtree(self.temp_dir_in)
        shutil.rmtree(self.temp_dir_out)

    def process_locale_dir(self, locale_dir, files):
        if self.domain == "djangojs":
            for file in files:
                # We need to copy the JS files first, as otherwise babel will
                # attempt to read package.json files in subdirs, such as
                # base/package.json
                in_path = urllib.parse.urljoin(
                    self.temp_dir_in + "/", file.dirpath
                )
                os.makedirs(in_path, exist_ok=True)
                in_file = urllib.parse.urljoin(in_path + "/", file.file)
                shutil.copy2(file.path, in_file)
                out_path = urllib.parse.urljoin(
                    self.temp_dir_out + "/", file.dirpath
                )
                file.dirpath = out_path
            os.chdir(".transpile/")
            out, err, status = popen_wrapper(
                [
                    "npm",
                    "run",
                    "babel-transform-template-literals",
                    "--",
                    "--out-dir",
                    self.temp_dir_out,
                    self.temp_dir_in,
                ]
            )
            os.chdir("../")

        super().process_locale_dir(locale_dir, files)

    def write_po_file(self, potfile, locale):
        if self.domain == "djangojs":
            with open(potfile, encoding="utf-8") as fp:
                msgs = fp.read()
            # Remove temp dir path info
            msgs = msgs.replace(self.temp_dir_out + "/", "")
            with open(potfile, "w", encoding="utf-8") as fp:
                fp.write(msgs)
        super().write_po_file(potfile, locale)



import os
from django.conf import settings as django_settings

"""
Application that contains main app settings
(Hosts, middleware, templates, installed apps)
Urls to route between applications,
statics that are shared among different django applications
Also contains helper for starting tornado server and management commands
"""


def read_version():
    with open(os.path.join(django_settings.SRC_PATH, "version.txt")) as f:
        return f.read().splitlines()[0]


__version__ = False


def get_version():
    global __version__
    if not __version__:
        __version__ = read_version()
    return __version__



from django.conf.urls import url

from . import views

urlpatterns = [
    url(
        "^delete_document_style/$",
        views.delete_document_style,
        name="delete_document_style",
    ),
    url(
        "^save_document_style/$",
        views.save_document_style,
        name="save_document_style",
    ),
    url(
        "^delete_export_template/$",
        views.delete_export_template,
        name="delete_export_template",
    ),
    url(
        "^save_export_template/$",
        views.save_export_template,
        name="save_export_template",
    ),
]



from typing import Tuple

from document.models import AccessRight, Document, DocumentTemplate


class SessionUserInfo:
    """
    Class for string information about users in session
    author: akorovin
    """

    def __init__(self, current_user):
        self.user = current_user
        self.is_owner = False
        self.access_rights = "read"
        self.document_id = 0
        self.access_rights = dict()
        self.path = ""
        self.path_object = None

    def create_doc(self, template_id):
        template = DocumentTemplate.objects.filter(id=int(template_id)).first()
        if template:
            document = Document.objects.create(
                owner_id=self.user.id, template=template
            )
        else:
            document = Document.objects.create(owner_id=self.user.id)
        return document

    def init_access(self, document_id) -> Tuple[Document, bool]:
        """
        Initializes access to document by id,
        :param document_id:
        :type document_id:
        :param current_user:
        :type current_user:
        :return: Returns document and bool value that user can access
        :rtype: tuple
        """
        can_access = False
        document = Document.objects.filter(id=int(document_id)).first()
        if document is None:
            return (False, False)
        else:
            self.document_id = document.id
            if document.owner == self.user:
                self.access_rights = "write"
                self.is_owner = True
                self.path = document.path
                self.path_object = document
                can_access = True
            else:
                self.is_owner = False
                access_right = AccessRight.objects.filter(
                    document=document, user=self.user
                ).first()
                if access_right:
                    self.access_rights = access_right.rights
                    self.path = access_right.path
                    self.path_object = access_right
                    can_access = True
        return (document, can_access)



from django.shortcuts import render
from django.views.decorators.csrf import ensure_csrf_cookie
from django.contrib.sites.shortcuts import get_current_site
from django.contrib.flatpages.models import FlatPage
from django.http import JsonResponse
from django.views.decorators.http import require_POST

from allauth.socialaccount.models import providers

from .decorators import ajax_required
from . import get_version


@ensure_csrf_cookie
def app(request):
    """
    Load a page controlled by the JavaScript app.
    Used all user facing pages after login.
    """
    return render(request, "app.html", {"version": get_version()})


def api_404(request):
    """
    Show a 404 error within the API.
    """
    return render(request, "api_404.html", status=404)


@ajax_required
@require_POST
def configuration(request):
    """
    Load the configuration options of the page that are request dependent.
    """
    socialaccount_providers = []
    for provider in providers.registry.get_list():
        socialaccount_providers.append(
            {
                "id": provider.id,
                "name": provider.name,
                "login_url": provider.get_login_url(request),
            }
        )
    response = {
        "language": request.LANGUAGE_CODE,
        "socialaccount_providers": socialaccount_providers,
    }
    if request.user.is_authenticated:
        response["user"] = {
            "id": request.user.id,
            "username": request.user.username,
            "first_name": request.user.first_name,
            "name": request.user.readable_name,
            "last_name": request.user.last_name,
            "avatar": request.user.avatar_url,
            "emails": [],
            "socialaccounts": [],
            "is_authenticated": True,
        }

        for emailaddress in request.user.emailaddress_set.all():
            email = {
                "address": emailaddress.email,
            }
            if emailaddress.primary:
                email["primary"] = True
            if emailaddress.verified:
                email["verified"] = True
            response["user"]["emails"].append(email)
        for account in request.user.socialaccount_set.all():
            try:
                provider_account = account.get_provider_account()
                response["user"]["socialaccounts"].append(
                    {
                        "id": account.id,
                        "provider": account.provider,
                        "name": provider_account.to_str(),
                    }
                )
            except KeyError:
                # Social account provider has been removed.
                pass
        response["user"]["waiting_invites"] = request.user.invites_to.exists()

    else:
        response["user"] = {"is_authenticated": False}
    return JsonResponse(response, status=200)


def manifest_json(request):
    """
    Load the manifest.json.
    """
    return render(request, "manifest.json")


# view is shown only in admin interface, so authentication is taken care of
def admin_console(request):
    """
    Load the admin console page.
    """
    return render(request, "admin/console.html")


@ajax_required
@require_POST
def flatpage(request):
    """
    Models: `flatpages.flatpages`
    Context:
        flatpage
            `flatpages.flatpages` object
    """
    response = {}
    status = 404
    url = request.POST["url"]
    site_id = get_current_site(request).id
    flatpage = FlatPage.objects.filter(url=url, sites=site_id).first()
    if flatpage:
        status = 200
        response["title"] = flatpage.title
        response["content"] = flatpage.content
    return JsonResponse(response, status=status)



from builtins import object
from django.contrib import auth
from django.conf import settings
from importlib import import_module


class DjangoHandlerMixin(object):
    def get_django_session(self):
        if not hasattr(self, "_session"):
            engine = import_module(settings.SESSION_ENGINE)
            session_key = self.get_cookie(settings.SESSION_COOKIE_NAME)
            self._session = engine.SessionStore(session_key)
        return self._session

    def get_current_user(self):
        # get_user needs a django request object, but only looks at the session

        class Dummy(object):
            pass

        django_request = Dummy()
        django_request.session = self.get_django_session()
        return auth.get_user(django_request)



#!/usr/bin/env python3
import os
import sys
from importlib import import_module

from django.db import connections


if "COVERAGE_PROCESS_START" in os.environ:
    import coverage

    coverage.process_startup()

SRC_PATH = os.path.dirname(os.path.realpath(__file__))
os.environ.setdefault("SRC_PATH", SRC_PATH)


# The following lines are needed until Tornado supports ASGI or Tornado has
# been replaced.
class Connections:
    pass


os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"
setattr(connections, "_connections", Connections)


def inner(default_project_path):
    sys.path.append(SRC_PATH)
    sys_argv = sys.argv
    PROJECT_PATH = False
    while "--pythonpath" in sys_argv:
        index = sys_argv.index("--pythonpath")
        PROJECT_PATH = os.path.join(os.getcwd(), sys_argv[index + 1])
        sys.path.insert(0, PROJECT_PATH)
        # We prevent the pythonpath to be handled later on by removing it from
        # sys_argv
        sys_argv = sys_argv[:index] + sys_argv[index + 2 :]
    if not PROJECT_PATH:
        PROJECT_PATH = default_project_path
        sys.path.insert(0, PROJECT_PATH)
    os.environ.setdefault("PROJECT_PATH", PROJECT_PATH)
    if "--settings" in sys_argv:
        index = sys_argv.index("--settings")
        SETTINGS_MODULE = sys_argv[index + 1]
    else:
        SETTINGS_MODULE = "configuration"
    mod = False
    # There are three levels of settings, each overiding the previous one:
    # global_settings.py, settings.py and configuration.py
    from django.conf import global_settings as CONFIGURATION
    from base import settings as SETTINGS

    SETTINGS_PATHS = [SETTINGS.__file__]
    for setting in dir(SETTINGS):
        setting_value = getattr(SETTINGS, setting)
        setattr(CONFIGURATION, setting, setting_value)
    try:
        mod = import_module(SETTINGS_MODULE)
    except ModuleNotFoundError:
        SETTINGS_MODULE = None
    if mod:
        SETTINGS_PATHS.append(mod.__file__)
        for setting in dir(mod):
            if setting.isupper():
                setattr(CONFIGURATION, setting, getattr(mod, setting))
    INSTALLED_APPS = CONFIGURATION.BASE_INSTALLED_APPS + list(
        CONFIGURATION.INSTALLED_APPS
    )
    for app in CONFIGURATION.REMOVED_APPS:
        INSTALLED_APPS.remove(app)
    from django.conf import settings

    settings.configure(
        CONFIGURATION,
        SETTINGS_MODULE=SETTINGS_MODULE,
        SETTINGS_PATHS=SETTINGS_PATHS,
        INSTALLED_APPS=INSTALLED_APPS,
        MIDDLEWARE=(
            CONFIGURATION.BASE_MIDDLEWARE + list(CONFIGURATION.MIDDLEWARE)
        ),
    )
    os.environ["TZ"] = settings.TIME_ZONE
    if sys_argv[1] in ["version", "--version"]:
        from base import get_version

        sys.stdout.write(get_version() + "\n")
        return
    from django.core.management import execute_from_command_line

    execute_from_command_line(sys_argv)


def entry():
    os.environ.setdefault("NO_COMPILEMESSAGES", "true")
    inner(os.getcwd())


if __name__ == "__main__":
    inner(SRC_PATH)



import uuid
import atexit
import logging
import copy
from time import mktime, time
from copy import deepcopy


from django.db.utils import DatabaseError
from django.db.models import F, Q
from django.conf import settings

from document.helpers.session_user_info import SessionUserInfo
from document import prosemirror
from document.helpers.serializers import PythonWithURLSerializer
from base.ws_handler import BaseWebSocketHandler
from document.models import (
    COMMENT_ONLY,
    CAN_UPDATE_DOCUMENT,
    CAN_COMMUNICATE,
    FW_DOCUMENT_VERSION,
    DocumentTemplate,
    Document,
)
from usermedia.models import Image, DocumentImage, UserImage

# settings_JSONPATCH
from jsonpatch import apply_patch, JsonPatchConflict, JsonPointerException
from tornado.escape import json_encode

# end settings_JSONPATCH

logger = logging.getLogger(__name__)


class WebSocket(BaseWebSocketHandler):
    sessions = dict()
    history_length = 1000  # Only keep the last 1000 diffs

    def open(self, arg):
        super().open(arg)
        if len(self.args) < 2:
            self.access_denied()
            return
        self.sessionument_id = int(self.args[0])
        logger.debug(
            f"Action:Document socket opened by user. "
            f"URL:{self.endpoint} User:{self.user.id} ParticipantID:{self.id}"
        )

    def confirm_diff(self, rid):
        response = {"type": "confirm_diff", "rid": rid}
        self.send_message(response)

    def subscribe_doc(self, connection_count=0):
        self.user_info = SessionUserInfo(self.user)
        doc_db, can_access = self.user_info.init_access(self.sessionument_id)
        if not can_access or float(doc_db.doc_version) != FW_DOCUMENT_VERSION:
            self.access_denied()
            return
        if (
            doc_db.id in WebSocket.sessions
            and len(WebSocket.sessions[doc_db.id]["participants"]) > 0
        ):
            logger.debug(
                f"Action:Serving already opened document. "
                f"URL:{self.endpoint} User:{self.user.id} "
                f" ParticipantID:{self.id}"
            )
            self.session = WebSocket.sessions[doc_db.id]
            self.id = max(self.session["participants"]) + 1
            self.session["participants"][self.id] = self
            template = False
        else:
            logger.debug(
                f"Action:Opening document from DB. "
                f"URL:{self.endpoint} User:{self.user.id} "
                f"ParticipantID:{self.id}"
            )
            self.id = 0
            if "type" in doc_db.content:
                content = doc_db.content
            else:
                content = copy.deepcopy(doc_db.template.content)
                if "type" not in content:
                    content["type"] = "article"
                if "content" not in content:
                    content["content"] = [{type: "title"}]
                doc_db.content = content
                doc_db.save()
            node = prosemirror.from_json({"type": "doc", "content": [content]})
            self.session = {
                "doc": doc_db,
                "node": node,
                "participants": {0: self},
                "last_saved_version": doc_db.version,
            }
            WebSocket.sessions[doc_db.id] = self.session
            if self.user_info.access_rights == "write":
                template = True
            else:
                template = False
        logger.debug(
            f"Action:Participant ID Assigned. URL:{self.endpoint} "
            f"User:{self.user.id} ParticipantID:{self.id}"
        )
        self.send_message({"type": "subscribed"})
        if connection_count < 1:
            self.send_styles()
            self.send_document(False, template)
        if connection_count >= 1:
            # If the user is reconnecting pass along access_rights to
            # front end to compare it with previous rights.
            self.send_message(
                {
                    "type": "access_right",
                    "access_right": self.user_info.access_rights,
                }
            )
        if self.can_communicate():
            self.handle_participant_update()

    def send_styles(self):
        doc_db = self.session["doc"]
        response = dict()
        response["type"] = "styles"
        serializer = PythonWithURLSerializer()
        export_temps = serializer.serialize(
            doc_db.template.exporttemplate_set.all(),
            fields=["file_type", "template_file", "title"],
        )
        document_styles = serializer.serialize(
            doc_db.template.documentstyle_set.all(),
            use_natural_foreign_keys=True,
            fields=["title", "slug", "contents", "documentstylefile_set"],
        )
        document_templates = {}
        for obj in DocumentTemplate.objects.filter(
            Q(user=self.user) | Q(user=None)
        ).order_by(F("user").desc(nulls_first=True)):
            document_templates[obj.import_id] = {
                "title": obj.title,
                "id": obj.id,
            }

        response["styles"] = {
            "export_templates": [obj["fields"] for obj in export_temps],
            "document_styles": [obj["fields"] for obj in document_styles],
            "document_templates": document_templates,
        }
        self.send_message(response)

    def unfixable(self):
        self.send_document()

    def send_document(self, messages=False, template=False):
        response = dict()
        response["type"] = "doc_data"
        doc_owner = self.session["doc"].owner
        response["doc_info"] = {
            "id": self.session["doc"].id,
            "is_owner": self.user_info.is_owner,
            "access_rights": self.user_info.access_rights,
            "path": self.user_info.path,
            "owner": {
                "id": doc_owner.id,
                "name": doc_owner.readable_name,
                "username": doc_owner.username,
                "avatar": doc_owner.avatar_url,
                "contacts": [],
            },
        }
        response["doc"] = {
            "v": self.session["doc"].version,
            "content": self.session["doc"].content,
            "bibliography": self.session["doc"].bibliography,
            "images": {},
        }
        if template:
            response["doc"]["template"] = {
                "id": self.session["doc"].template.id,
                "content": self.session["doc"].template.content,
            }
        if messages:
            response["m"] = messages
        response["time"] = int(time()) * 1000
        for dimage in DocumentImage.objects.filter(
            document_id=self.session["doc"].id
        ):
            image = dimage.image
            field_obj = {
                "id": image.id,
                "title": dimage.title,
                "copyright": dimage.copyright,
                "image": image.image.url,
                "file_type": image.file_type,
                "added": mktime(image.added.timetuple()) * 1000,
                "checksum": image.checksum,
                "cats": [],
            }
            if image.thumbnail:
                field_obj["thumbnail"] = image.thumbnail.url
                field_obj["height"] = image.height
                field_obj["width"] = image.width
            response["doc"]["images"][image.id] = field_obj
        if self.user_info.access_rights == "read-without-comments":
            response["doc"]["comments"] = []
        elif self.user_info.access_rights in ["review", "review-tracked"]:
            # Reviewer should only get his/her own comments
            filtered_comments = {}
            for key, value in list(self.session["doc"].comments.items()):
                if value["user"] == self.user_info.user.id:
                    filtered_comments[key] = value
            response["doc"]["comments"] = filtered_comments
        else:
            response["doc"]["comments"] = self.session["doc"].comments
        for contact in doc_owner.contacts.all():
            contact_object = {
                "id": contact.id,
                "name": contact.readable_name,
                "username": contact.get_username(),
                "avatar": contact.avatar_url,
                "type": "user",
            }
            response["doc_info"]["owner"]["contacts"].append(contact_object)
        if self.user_info.is_owner:
            for contact in doc_owner.invites_by.all():
                contact_object = {
                    "id": contact.id,
                    "name": contact.username,
                    "username": contact.username,
                    "avatar": contact.avatar_url,
                    "type": "userinvite",
                }
                response["doc_info"]["owner"]["contacts"].append(
                    contact_object
                )
        response["doc_info"]["session_id"] = self.id
        self.send_message(response)

    def reject_message(self, message):
        if message["type"] == "diff":
            self.send_message({"type": "reject_diff", "rid": message["rid"]})

    def handle_message(self, message):
        if message["type"] == "subscribe":
            connection_count = 0
            if "connection" in message:
                connection_count = message["connection"]
            self.subscribe_doc(connection_count)
            return
        if self.user_info.document_id not in WebSocket.sessions:
            logger.debug(
                f"Action:Receiving message for closed document. "
                f"URL:{self.endpoint} User:{self.user.id} "
                f"ParticipantID:{self.id}"
            )
            return
        if message["type"] == "get_document":
            self.send_document()
        elif (
            message["type"] == "participant_update" and self.can_communicate()
        ):
            self.handle_participant_update()
        elif message["type"] == "chat" and self.can_communicate():
            self.handle_chat(message)
        elif message["type"] == "check_version":
            self.check_version(message)
        elif message["type"] == "selection_change":
            self.handle_selection_change(message)
        elif message["type"] == "diff" and self.can_update_document():
            self.handle_diff(message)
        elif message["type"] == "path_change":
            self.handle_path_change(message)

    def update_bibliography(self, bibliography_updates):
        for bu in bibliography_updates:
            if "id" not in bu:
                continue
            id = bu["id"]
            if bu["type"] == "update":
                self.session["doc"].bibliography[id] = bu["reference"]
            elif bu["type"] == "delete":
                del self.session["doc"].bibliography[id]

    def update_images(self, image_updates):
        for iu in image_updates:
            if "id" not in iu:
                continue
            id = iu["id"]
            if iu["type"] == "update":
                # Ensure that access rights exist
                if not UserImage.objects.filter(
                    image__id=id, owner=self.user_info.user
                ).exists():
                    continue
                doc_image = DocumentImage.objects.filter(
                    document_id=self.session["doc"].id, image_id=id
                ).first()
                if doc_image:
                    doc_image.title = iu["image"]["title"]
                    doc_image.copyright = iu["image"]["copyright"]
                    doc_image.save()
                else:
                    DocumentImage.objects.create(
                        document_id=self.session["doc"].id,
                        image_id=id,
                        title=iu["image"]["title"],
                        copyright=iu["image"]["copyright"],
                    )
            elif iu["type"] == "delete":
                DocumentImage.objects.filter(
                    document_id=self.session["doc"].id, image_id=id
                ).delete()
                for image in Image.objects.filter(id=id):
                    if image.is_deletable():
                        image.delete()

    def update_comments(self, comments_updates):
        comments_updates = deepcopy(comments_updates)
        for cd in comments_updates:
            if "id" not in cd:
                # ignore
                continue
            id = cd["id"]
            if cd["type"] == "create":
                self.session["doc"].comments[id] = {
                    "user": cd["user"],
                    "username": cd["username"],
                    "assignedUser": cd["assignedUser"],
                    "assignedUsername": cd["assignedUsername"],
                    "date": cd["date"],
                    "comment": cd["comment"],
                    "isMajor": cd["isMajor"],
                    "resolved": cd["resolved"],
                }
            elif cd["type"] == "delete":
                del self.session["doc"].comments[id]
            elif cd["type"] == "update":
                self.session["doc"].comments[id]["comment"] = cd["comment"]
                if "isMajor" in cd:
                    self.session["doc"].comments[id]["isMajor"] = cd["isMajor"]
                if "assignedUser" in cd and "assignedUsername" in cd:
                    self.session["doc"].comments[id]["assignedUser"] = cd[
                        "assignedUser"
                    ]
                    self.session["doc"].comments[id]["assignedUsername"] = cd[
                        "assignedUsername"
                    ]
                if "resolved" in cd:
                    self.session["doc"].comments[id]["resolved"] = cd[
                        "resolved"
                    ]
            elif cd["type"] == "add_answer":
                if "answers" not in self.session["doc"].comments[id]:
                    self.session["doc"].comments[id]["answers"] = []
                self.session["doc"].comments[id]["answers"].append(
                    {
                        "id": cd["answerId"],
                        "user": cd["user"],
                        "username": cd["username"],
                        "date": cd["date"],
                        "answer": cd["answer"],
                    }
                )
            elif cd["type"] == "delete_answer":
                answer_id = cd["answerId"]
                for answer in self.session["doc"].comments[id]["answers"]:
                    if answer["id"] == answer_id:
                        self.session["doc"].comments[id]["answers"].remove(
                            answer
                        )
            elif cd["type"] == "update_answer":
                answer_id = cd["answerId"]
                for answer in self.session["doc"].comments[id]["answers"]:
                    if answer["id"] == answer_id:
                        answer["answer"] = cd["answer"]

    def handle_participant_update(self):
        WebSocket.send_participant_list(self.user_info.document_id)

    def handle_chat(self, message):
        chat = {
            "id": str(uuid.uuid4()),
            "body": message["body"],
            "from": self.user_info.user.id,
            "type": "chat",
        }
        WebSocket.send_updates(chat, self.user_info.document_id)

    def handle_selection_change(self, message):
        if (
            self.user_info.document_id in WebSocket.sessions
            and message["v"] == self.session["doc"].version
        ):
            WebSocket.send_updates(
                message,
                self.user_info.document_id,
                self.id,
                self.user_info.user.id,
            )

    def handle_path_change(self, message):
        if (
            self.user_info.document_id in WebSocket.sessions
            and self.user_info.path_object
        ):
            self.user_info.path_object.path = message["path"]
            self.user_info.path_object.save(
                update_fields=[
                    "path",
                ]
            )
            WebSocket.send_updates(
                message,
                self.user_info.document_id,
                self.id,
                self.user_info.user.id,
            )

    # Checks if the diff only contains changes to comments.
    def only_comments(self, message):
        allowed_operations = ["addMark", "removeMark"]
        only_comment = True
        if "ds" in message:  # ds = document steps
            for step in message["ds"]:
                if not (
                    step["stepType"] in allowed_operations
                    and step["mark"]["type"] == "comment"
                ):
                    only_comment = False
        return only_comment

    def handle_diff(self, message):
        pv = message["v"]
        dv = self.session["doc"].version
        logger.debug(
            f"Action:Handling Diff. URL:{self.endpoint} User:{self.user.id} "
            f"ParticipantID:{self.id} Client version:{pv} "
            f"Server version:{dv} Message:{message}"
        )
        if (
            self.user_info.access_rights in COMMENT_ONLY
            and not self.only_comments(message)
        ):
            logger.error(
                f"Action:Received non-comment diff from comment-only "
                f"collaborator.Discarding URL:{self.endpoint} "
                f"User:{self.user.id} ParticipantID:{self.id}"
            )
            return
        if pv == dv:
            if settings.JSONPATCH:
                if "jd" in message:  # jd = json diff
                    backup = False
                    if len(message["jd"]) > 1:
                        # There is more than one patch operation so if the
                        # patch fails, we might already have applied a previous
                        # patch operation. Therefore we take a backup now so
                        # that we can roll back if needed.
                        backup = deepcopy(self.session["doc"].content)
                    try:
                        apply_patch(
                            self.session["doc"].content, message["jd"], True
                        )
                    except (JsonPatchConflict, JsonPointerException):
                        if backup:
                            self.session["doc"].content = backup
                        logger.exception(
                            f"Action:Cannot apply json diff. "
                            f"URL:{self.endpoint} User:{self.user.id} "
                            f"ParticipantID:{self.id}"
                        )
                        logger.error(
                            f"Action:Patch Exception URL:{self.endpoint} "
                            f"User:{self.user.id} ParticipantID:{self.id} "
                            f"Message:{json_encode(message)}"
                        )
                        logger.error(
                            f"Action:Patch Exception URL:{self.endpoint} "
                            f"User:{self.user.id} ParticipantID:{self.id} "
                            f"Document:"
                            f"{json_encode(self.session['doc'].content)}"
                        )
                        self.unfixable()
                        patch_msg = {
                            "type": "patch_error",
                            "user_id": self.user.id,
                        }
                        self.send_message(patch_msg)
                        # Reset collaboration to avoid any data loss issues.
                        self.reset_collaboration(
                            patch_msg, self.user_info.document_id, self.id
                        )
                        return
                    # The json diff is only needed by the python backend which
                    # does not understand the steps. It can therefore be
                    # removed before broadcast to other clients.
                    del message["jd"]
            elif "ds" in message:  # ds = document steps
                updated_node = prosemirror.apply(
                    message["ds"], self.session["node"]
                )
                if updated_node:
                    self.session["node"] = updated_node
                else:
                    self.unfixable()
                    patch_msg = {
                        "type": "patch_error",
                        "user_id": self.user.id,
                    }
                    self.send_message(patch_msg)
                    # Reset collaboration to avoid any data loss issues.
                    self.reset_collaboration(
                        patch_msg, self.user_info.document_id, self.id
                    )
                    return
                self.session["doc"].content = prosemirror.to_mini_json(
                    self.session["node"].first_child
                )
            self.session["doc"].diffs.append(message)
            self.session["doc"].diffs = self.session["doc"].diffs[
                -self.history_length :
            ]
            self.session["doc"].version += 1
            if "ti" in message:  # ti = title
                self.session["doc"].title = message["ti"][-255:]
            if "cu" in message:  # cu = comment updates
                self.update_comments(message["cu"])
            if "bu" in message:  # bu = bibliography updates
                self.update_bibliography(message["bu"])
            if "iu" in message:  # iu = image updates
                self.update_images(message["iu"])
            if self.session["doc"].version % settings.DOC_SAVE_INTERVAL == 0:
                WebSocket.save_document(self.user_info.document_id)
            self.confirm_diff(message["rid"])
            WebSocket.send_updates(
                message,
                self.user_info.document_id,
                self.id,
                self.user_info.user.id,
            )
        elif pv < dv:
            if pv + len(self.session["doc"].diffs) >= dv:
                # We have enough diffs stored to fix it.
                number_diffs = pv - dv
                logger.debug(
                    f"Action:Resending document diffs. URL:{self.endpoint} "
                    f"User:{self.user.id} ParticipantID:{self.id} "
                    f"number of messages to be resent:{number_diffs}"
                )
                messages = self.session["doc"].diffs[number_diffs:]
                for message in messages:
                    new_message = message.copy()
                    new_message["server_fix"] = True
                    self.send_message(new_message)
            else:
                logger.debug(
                    f"Action:User is on a very old version of the document. "
                    f"URL:{self.endpoint} User:{self.user.id} "
                    f"ParticipantID:{self.id}"
                )
                # Client has a version that is too old to be fixed
                self.unfixable()
                return
        else:
            # Client has a higher version than server. Something is fishy!
            logger.debug(
                f"Action:User has higher document version than server.Fishy! "
                f"URL:{self.endpoint} User:{self.user.id} "
                f"ParticipantID:{self.id}"
            )

    def check_version(self, message):
        pv = message["v"]
        dv = self.session["doc"].version
        logger.debug(
            f"Action:Checking version of document. URL:{self.endpoint} "
            f"User:{self.user.id} ParticipantID:{self.id} "
            f"Client document version:{pv} Server document version:{dv}"
        )
        if pv == dv:
            response = {
                "type": "confirm_version",
                "v": pv,
            }
            self.send_message(response)
            return
        elif pv + len(self.session["doc"].diffs) >= dv:
            number_diffs = pv - dv
            logger.debug(
                f"Action:Resending document diffs. URL:{self.endpoint} "
                f"User:{self.user.id} ParticipantID:{self.id}"
                f"number of messages to be resent:{number_diffs}"
            )
            messages = self.session["doc"].diffs[number_diffs:]
            self.send_document(messages)
            return
        else:
            logger.debug(
                f"Action:User is on a very old version of the document. "
                f"URL:{self.endpoint} User:{self.user.id} "
                f"ParticipantID:{self.id}"
            )
            # Client has a version that is too old
            self.unfixable()
            return

    def can_update_document(self):
        return self.user_info.access_rights in CAN_UPDATE_DOCUMENT

    def can_communicate(self):
        return self.user_info.access_rights in CAN_COMMUNICATE

    def on_close(self):
        logger.debug(
            f"Action:Closing websocket. URL:{self.endpoint} "
            f"User:{self.user.id} ParticipantID:{self.id}"
        )
        if (
            hasattr(self, "session")
            and hasattr(self, "user_info")
            and hasattr(self.user_info, "document_id")
            and self.user_info.document_id in WebSocket.sessions
            and hasattr(self, "id")
            and self.id
            in WebSocket.sessions[self.user_info.document_id]["participants"]
        ):
            del self.session["participants"][self.id]
            if len(self.session["participants"]) == 0:
                WebSocket.save_document(self.user_info.document_id)
                del WebSocket.sessions[self.user_info.document_id]
                logger.debug(
                    f"Action:No participants for the document. "
                    f"URL:{self.endpoint} User:{self.user.id}"
                )
            else:
                WebSocket.send_participant_list(self.user_info.document_id)

    @classmethod
    def send_participant_list(cls, document_id):
        if document_id in WebSocket.sessions:
            participant_list = []
            for session_id, waiter in list(
                cls.sessions[document_id]["participants"].items()
            ):
                access_rights = waiter.user_info.access_rights
                if access_rights not in CAN_COMMUNICATE:
                    continue
                participant_list.append(
                    {
                        "session_id": session_id,
                        "id": waiter.user_info.user.id,
                        "name": waiter.user_info.user.readable_name,
                        "avatar": waiter.user_info.user.avatar_url,
                    }
                )
            message = {
                "participant_list": participant_list,
                "type": "connections",
            }
            WebSocket.send_updates(message, document_id)

    @classmethod
    def reset_collaboration(cls, patch_exception_msg, document_id, sender_id):
        logger.debug(
            f"Action:Resetting collaboration. DocumentID:{document_id} "
            f"Patch conflict triggered. ParticipantID:{sender_id} "
            f"waiters:{len(cls.sessions[document_id]['participants'])}"
        )
        for waiter in list(cls.sessions[document_id]["participants"].values()):
            if waiter.id != sender_id:
                waiter.unfixable()
                waiter.send_message(patch_exception_msg)

    @classmethod
    def send_updates(cls, message, document_id, sender_id=None, user_id=None):
        logger.debug(
            f"Action:Sending message to waiters. DocumentID:{document_id} "
            f"waiters:{len(cls.sessions[document_id]['participants'])}"
        )
        for waiter in list(cls.sessions[document_id]["participants"].values()):
            if waiter.id != sender_id:
                access_rights = waiter.user_info.access_rights
                if "comments" in message and len(message["comments"]) > 0:
                    # Filter comments if needed
                    if access_rights == "read-without-comments":
                        # The reader should not receive the comments update, so
                        # we remove the comments from the copy of the message
                        # sent to the reviewer. We still need to send the rest
                        # of the message as it may contain other diff
                        # information.
                        message = deepcopy(message)
                        message["comments"] = []
                    elif (
                        access_rights in ["review", "review-tracked"]
                        and user_id != waiter.user_info.user.id
                    ):
                        # The reviewer should not receive comments updates from
                        # others than themselves, so we remove the comments
                        # from the copy of the message sent to the reviewer
                        # that are not from them. We still need to send the
                        # rest of the message as it may contain other diff
                        # information.
                        message = deepcopy(message)
                        message["comments"] = []
                elif (
                    message["type"] in ["chat", "connections"]
                    and access_rights not in CAN_COMMUNICATE
                ):
                    continue
                elif (
                    message["type"] == "selection_change"
                    and access_rights not in CAN_COMMUNICATE
                    and user_id != waiter.user_info.user.id
                ):
                    continue
                elif (
                    message["type"] == "path_change"
                    and user_id != waiter.user_info.user.id
                ):
                    continue
                waiter.send_message(message)

    @classmethod
    def save_document(cls, document_id):
        session = cls.sessions[document_id]
        if session["doc"].version == session["last_saved_version"]:
            return
        logger.debug(
            f"Action:Saving document to DB. DocumentID:{session['doc'].id} "
            f"Doc version:{session['doc'].version}"
        )
        try:
            # this try block is to avoid a db exception
            # in case the doc has been deleted from the db
            # in fiduswriter the owner of a doc could delete a doc
            # while an invited writer is editing the same doc
            session["doc"].save(
                update_fields=[
                    "title",
                    "version",
                    "content",
                    "diffs",
                    "comments",
                    "bibliography",
                    "updated",
                ]
            )

        except DatabaseError as e:
            expected_msg = "Save with update_fields did not affect any rows."
            if str(e) == expected_msg:
                cls.__insert_document(doc=session["doc"])
            else:
                raise e
        session["last_saved_version"] = session["doc"].version

    @classmethod
    def __insert_document(cls, doc: Document) -> None:
        """
        Purpose:
        during plugin tests we experienced Integrity errors
         at the end of tests.
        This exception occurs while handling another exception,
         so in order to have a clean tests output
          we raise the exception in a way we don't output
           misleading error messages related to different exceptions

        :param doc: socket document model instance
        :return: None
        """
        from django.db.utils import IntegrityError

        try:
            doc.save()
        except IntegrityError as e:
            raise IntegrityError(
                "plugin test error when we try to save a doc already deleted "
                "along with the rest of db data so it "
                "raises an Integrity error: {}".format(e)
            ) from None

    @classmethod
    def save_all_docs(cls):
        for document_id in cls.sessions:
            cls.save_document(document_id)


atexit.register(WebSocket.save_all_docs)



import shutil
import os
import json
from subprocess import call

from django.apps import apps
from base.management import BaseCommand
from django.core.management import call_command
from django.conf import settings


class Command(BaseCommand):
    help = "Check JavaScript files with ESLint"

    def add_arguments(self, parser):
        parser.add_argument(
            "dir", nargs="?", default=".", help="Directory to check"
        )
        parser.add_argument(
            "--fix",
            action="store_true",
            dest="fix",
            default=False,
            help="Whether to also attempt to fix files",
        )

    def handle(self, *args, **options):
        self.stdout.write("Linting JavaScript code...")
        call_command("npm_install")
        shutil.os.chdir(settings.PROJECT_PATH)
        apps_paths = []
        for app in list(apps.get_app_configs()):
            apps_paths.append(app.path)
        os.environ["DJANGO_APPS_PATHS"] = json.dumps(apps_paths)
        command_array = [
            os.path.join(
                settings.PROJECT_PATH, ".transpile/node_modules/.bin/eslint"
            ),
            "--max-warnings=0",
            "--ignore-path",
            os.path.join(settings.SRC_PATH, ".eslintignore"),
            "-c",
            os.path.join(settings.SRC_PATH, ".eslintrc.js"),
            options["dir"],
        ]
        if options["fix"]:
            command_array.append("--fix")
        return_value = call(command_array)
        if return_value > 0:
            exit(return_value)



from importlib import import_module

from django.conf import settings
from django.core.wsgi import get_wsgi_application

from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.web import Application, FallbackHandler, StaticFileHandler
from tornado.wsgi import WSGIContainer

from base.handlers import DjangoStaticFilesHandler, HelloHandler, RobotsHandler


def make_tornado_server():
    wsgi_app = WSGIContainer(get_wsgi_application())
    tornado_url_list = [
        (
            r"/static/(.*)",
            DjangoStaticFilesHandler,
            {"default_filename": "none.img"},
        ),
        (r"/media/(.*)", StaticFileHandler, {"path": settings.MEDIA_ROOT}),
        ("/hello-tornado", HelloHandler),
        ("/robots.txt", RobotsHandler),
    ]

    for app in settings.INSTALLED_APPS:
        app_name = app.rsplit(".", 1).pop()
        # add proxy views
        try:
            proxy_module = import_module("%s.proxy_views" % app)
        except ImportError:
            pass
        else:
            tornado_url_list += [
                ("/proxy/%s/([^?]*)" % app_name, proxy_module.Proxy)
            ]
        # add ws views
        try:
            ws_module = import_module("%s.ws_views" % app)
        except ImportError:
            pass
        else:
            tornado_url_list += [
                (
                    "/ws/%s/([^?]*)" % app_name,
                    ws_module.WebSocket,
                    dict(app_name=app_name),
                )
            ]
    tornado_url_list += [(".*", FallbackHandler, dict(fallback=wsgi_app))]
    tornado_app = Application(
        tornado_url_list,
        debug=settings.DEBUG,
        websocket_ping_interval=settings.WEBSOCKET_PING_INTERVAL,
        compress_response=True,
    )
    server = HTTPServer(tornado_app)
    server.xheaders = True
    return server


def run(port):
    make_tornado_server().listen(int(port))
    IOLoop.current().start()



from django.core.management import call_command
from npm_mjs.management.commands.collectstatic import (
    Command as CollectStaticCommand,
)

from base.management import BaseCommand


class Command(CollectStaticCommand, BaseCommand):
    def handle(self, *args, **options):
        call_command("transpile")
        return super().handle(*args, **options)



# Generated by Django 3.1.4 on 2021-06-04 05:58

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ("user", "0007_userinvite_to"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userinvite",
            name="key",
            field=models.UUIDField(default=uuid.uuid4, unique=True),
        ),
    ]



# Generated by Django 3.1.4 on 2021-04-02 18:49

from django.db import migrations


def teammembers_to_contacts(apps, schema_editor):
    TeamMember = apps.get_model("user", "TeamMember")
    team_members = TeamMember.objects.all().iterator()
    for team_member in team_members:
        member = team_member.member
        leader = team_member.leader
        leader.contacts.add(member)
        member.contacts.add(leader)
    TeamMember.objects.all().delete()


def contacts_to_teammembers(apps, schema_editor):
    User = apps.get_model("user", "User")
    TeamMember = apps.get_model("user", "TeamMember")
    users = User.objects.all().iterator()
    for user in users:
        for contact in user.contacts.all():
            TeamMember.objects.get_or_create(leader=user, member=contact)
        user.contacts.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ("user", "0003_auto_20210411_1440"),
    ]

    operations = [
        migrations.RunPython(teammembers_to_contacts, contacts_to_teammembers),
    ]
